# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import sys

import click

from aiida.backends.utils import load_dbenv, is_dbenv_loaded
from aiida.cmdline import delayed_load_node as load_node
from aiida.cmdline.baseclass import (
    VerdiCommandRouter, VerdiCommandWithSubcommands)
from aiida.cmdline.commands.node import _Label, _Description
from aiida.common.exceptions import MultipleObjectsError
from aiida.cmdline.commands import verdi
from aiida.plugins.entry_point import get_entry_point_names


class Data(VerdiCommandRouter):
    """
    Setup and manage data specific types

    There is a list of subcommands for managing specific types of data.
    For instance, 'data upf' manages pseudopotentials in the UPF format.
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        ## Add here the classes to be supported.
        self.routed_subcommands = {
            'upf': _Upf,
            'structure': _Structure,
            'bands': _Bands,
            'cif': _Cif,
            'trajectory': _Trajectory,
            'parameter': _Parameter,
            'array': _Array,
            'label': _Label,
            'remote': _Remote,
            'description': _Description,
        }
        entry_point_group = 'aiida.cmdline.data'
        for entry_point_name in get_entry_point_names(entry_point_group):
            self.routed_subcommands[entry_point_name] = verdi

class Listable(object):
    """
    Provides shell completion for listable data nodes.

    .. note:: classes, inheriting Listable, MUST define value for property
        ``dataclass`` (preferably in ``__init__``), which
        has to point to correct \*Data class.
    """

    def list(self, *args):
        """
        List all instances of given data class.

        :param args: a list of command line arguments.
        """
        import argparse

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='List {} objects.'.format(self.dataclass.__name__))

        self.append_list_cmdline_arguments(parser)

        parser.add_argument('--vseparator', default="\t",
                            help="specify vertical separator for fields. "
                                 "Default '\\t'.",
                            type=str, action='store')
        parser.add_argument('--header', default=True,
                            help="print a header with column names. "
                                 "Default option.",
                            dest="header", action='store_true')
        parser.add_argument('--no-header', '-H',
                            help="do not print a header with column names.",
                            dest="header", action='store_false')

        args = list(args)
        parsed_args = parser.parse_args(args)

        # print "parsed_args ===>", parsed_args

        entry_list = self.query(parsed_args)

        vsep = parsed_args.vseparator
        if entry_list:
            to_print = ""
            if parsed_args.header:
                to_print += vsep.join(self.get_column_names()) + "\n"
            for entry in sorted(entry_list, key=lambda x: int(x[0])):
                to_print += vsep.join(entry) + "\n"
            sys.stdout.write(to_print)

    def query(self, args):
        """
        Perform the query
        """
        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.orm.querybuilder import QueryBuilder
        from aiida.orm.implementation import Group
        from aiida.orm.user import User
        from aiida.orm.backend import construct_backend

        backend = construct_backend()

        qb = QueryBuilder()
        if args.all_users is False:
            user = backend.users.get_automatic_user()
            qb.append(User, tag="creator", filters={"email": user.email})
        else:
            qb.append(User, tag="creator")

        data_filters = dict()
        self.query_past_days(data_filters, args)
        qb.append(self.dataclass, tag="data", created_by="creator",
                  filters=data_filters, project=["id"])

        group_filters = {}
        self.query_group(group_filters, args)
        if group_filters:
            qb.append(Group, tag="group", filters=group_filters,
                      group_of="data")

        qb.order_by({self.dataclass: {'ctime': 'asc'}})

        object_list = qb.distinct()

        entry_list = []
        for [id] in object_list.all():
            entry_list.append([str(id)])
        return entry_list

    def query_past_days(self, filters, args):
        """
        Subselect to filter data nodes by their age.

        :param filters: the filters to be enriched.
        :param args: a namespace with parsed command line parameters.
        """
        from aiida.utils import timezone
        import datetime
        if args.past_days is not None:
            now = timezone.now()
            n_days_ago = now - datetime.timedelta(days=args.past_days)
            filters.update({"ctime": {'>=': n_days_ago}})
        return filters

    def query_group(self, filters, args):
        """
        Subselect to filter data nodes by their group.

        :param q_object: a query object
        :param args: a namespace with parsed command line parameters.
        """
        if args.group_name is not None:
            filters.update({"name": {"in": args.group_name}})
        if args.group_pk is not None:
            filters.update({"id": {"in": args.group_pk}})

    def append_list_cmdline_arguments(self, parser):
        """
        Append additional command line parameters, that are later parsed and
        used in the query construction.

        :param parser: instance of argparse.ArgumentParser
        """
        parser.add_argument('-p', '--past-days', metavar='N',
                            help="add a filter to show only objects created in the past N days",
                            type=int, action='store')
        parser.add_argument('-g', '--group-name', metavar='N', nargs="+", default=None,
                            help="add a filter to show only objects belonging to groups",
                            type=str, action='store')
        parser.add_argument('-G', '--group-pk', metavar='N', nargs="+", default=None,
                            help="add a filter to show only objects belonging to groups",
                            type=int, action='store')
        parser.add_argument('-A', '--all-users', action='store_true', default=False,
                            help="show objects for all users, rather than only for the"
                                 "current user")

    def get_column_names(self):
        """
        Return the list with column names.

        .. note:: neither the number nor correspondence of column names and
            actual columns in the output from the :py:meth:`query` are checked.
        """
        return ["ID"]


class Visualizable(object):
    """
    Provides shell completion for visualizable data nodes.

    .. note:: classes, inheriting Visualizable, MUST NOT contain
        attributes, starting with ``_show_``, which are not plugins for
        visualization.

    In order to specify a default visualization format, one has to override
    ``_default_show_format`` property (preferably in
    ``__init__``), setting it to the name of default visualization tool.
    """
    show_prefix = '_show_'
    show_parameters_postfix = '_parameters'

    def get_show_plugins(self):
        """
        Get the list of all implemented plugins for visualizing the structure.
        """
        method_names = dir(self)  # get list of class methods names
        valid_formats = [i[len(self.show_prefix):] for i in method_names
                         if i.startswith(self.show_prefix) and \
                         not i.endswith(self.show_parameters_postfix)]  # filter

        return {k: getattr(self, self.show_prefix + k) for k in valid_formats}

    def show(self, *args):
        """
        Show the data node with a visualization program.
        """
        # DEVELOPER NOTE: to add a new plugin, just add a _show_xxx() method.
        import argparse, os

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Visualize data object.')
        parser.add_argument('data_id', type=int, default=None, nargs="+",
                            help="ID of the data object to be visualized.")

        default_format = None
        try:
            default_format = self._default_show_format
        except AttributeError:
            if len(self.get_show_plugins().keys()) == 1:
                default_format = self.get_show_plugins().keys()[0]
            else:
                default_format = None

        parser.add_argument('--format', '-f', type=str, default=default_format,
                            help="Type of the visualization format/tool.",
                            choices=sorted(self.get_show_plugins().keys()))

        # Augmenting the command line parameters with ones, that are used by
        # individual plugins
        for cmd in dir(self):
            if not cmd.startswith(self.show_prefix) or \
                    not cmd.endswith(self.show_parameters_postfix):
                continue
            getattr(self, cmd)(parser)

        args = list(args)
        parsed_args = vars(parser.parse_args(args))

        data_id = parsed_args.pop('data_id')
        format = parsed_args.pop('format')

        # Removing the keys, whose values are None
        for key in parsed_args.keys():
            if parsed_args[key] is None:
                parsed_args.pop(key)

        if format is None:
            print >> sys.stderr, (
                "Default format is not defined, please specify.\n"
                "Valid formats are:")
            for i in self.get_show_plugins().keys():
                print >> sys.stderr, "  {}".format(i)
            sys.exit(1)

        # I can give in input the whole path to executable
        code_name = os.path.split(format)[-1]

        try:
            func = self.get_show_plugins()[code_name]
        except KeyError:
            print >> sys.stderr, "Not implemented; implemented plugins are:"
            print >> sys.stderr, "{}.".format(
                ",".join(self.get_show_plugins()))
            sys.exit(1)

        if not is_dbenv_loaded():
            load_dbenv()

        n_list = [load_node(id) for id in data_id]

        for n in n_list:
            try:
                if not isinstance(n, self.dataclass):
                    print >> sys.stderr, ("Node {} is of class {} instead "
                                          "of {}".format(n, type(n), self.dataclass))
                    sys.exit(1)
            except AttributeError:
                pass

        try:
            func(format, n_list, **parsed_args)
        except MultipleObjectsError:
            print >> sys.stderr, (
                "Visualization of multiple objects is not implemented "
                "for '{}'".format(format))
            sys.exit(1)


class Exportable(object):
    """
    Provides shell completion for exportable data nodes.

    .. note:: classes, inheriting Exportable, MUST NOT contain attributes,
        starting with ``_export_``, which are not plugins for exporting.
    """
    export_prefix = '_export_'
    export_parameters_postfix = '_parameters'

    def append_export_cmdline_arguments(self, parser):
        """
        Function (to be overloaded in a subclass) to add custom export command
        line arguments.

        :param parser: a ArgParse parser object
        :return: change the parser in place
        """
        pass

    def get_export_plugins(self):
        """
        Get the list of all implemented exporters for data class.
        """
        method_names = dir(self)  # get list of class methods names
        valid_formats = [i[len(self.export_prefix):] for i in method_names
                         if i.startswith(self.export_prefix) and \
                         not i.endswith(self.export_parameters_postfix)]  # filter

        return {k: getattr(self, self.export_prefix + k) for k in valid_formats}

    def export(self, *args):
        """
        Export the data node to a given format.
        """
        # DEVELOPER NOTE: to add a new plugin, just add a _export_xxx() method.
        import argparse

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Export data object.')
        parser.add_argument('-o', '--output', type=str, default='',
                            help="If present, store the output directly on a file "
                                 "with the given name. It is essential to use this option "
                                 "if more than one file needs to be created.")
        parser.add_argument('-y', '--overwrite', action='store_true',
                            help="If passed, overwrite files without checking.")
        parser.add_argument('data_id', type=int, default=None,
                            help="ID of the data object to be visualized.")

        self.append_export_cmdline_arguments(parser)

        default_format = None
        try:
            default_format = self._default_export_format
        except AttributeError:
            if len(self.get_export_plugins().keys()) == 1:
                default_format = self.get_export_plugins().keys()[0]
            else:
                default_format = None

        parser.add_argument('--format', '-f', type=str, default=default_format,
                            help="Type of the exported file.",
                            choices=sorted(self.get_export_plugins().keys()))

        # Augmenting the command line parameters with ones, that are used by
        # individual plugins
        for cmd in dir(self):
            if not cmd.startswith(self.export_prefix) or \
                    not cmd.endswith(self.export_parameters_postfix):
                continue
            getattr(self, cmd)(parser)

        args = list(args)
        parsed_args = vars(parser.parse_args(args))

        format = parsed_args.pop('format')
        data_id = parsed_args.pop('data_id')

        # Removing the keys, whose values are None
        for key in parsed_args.keys():
            if parsed_args[key] is None:
                parsed_args.pop(key)

        if format is None:
            print >> sys.stderr, (
                "Default format is not defined, please specify.\n"
                "Valid formats are:")
            for i in sorted(self.get_export_plugins().keys()):
                print >> sys.stderr, "  {}".format(i)
            sys.exit(1)

        output_fname = parsed_args.pop('output')
        if not output_fname:
            output_fname = ""

        overwrite = parsed_args.pop('overwrite')

        # if parsed_args:
        #    raise InternalError(
        #        "Some command line parameters were not properly parsed: {}".format(
        #            parsed_args.keys()
        #        ))

        try:
            func = self.get_export_plugins()[format]
        except KeyError:
            print >> sys.stderr, "Not implemented; implemented plugins are:"
            print >> sys.stderr, "{}.".format(
                ",".join(self.get_export_plugins()))
            sys.exit(1)

        if not is_dbenv_loaded():
            load_dbenv()

        n = load_node(data_id)

        try:
            if not isinstance(n, self.dataclass):
                print >> sys.stderr, ("Node {} is of class {} instead "
                                      "of {}".format(n, type(n), self.dataclass))
                sys.exit(1)
        except AttributeError:
            pass

        func(n, output_fname=output_fname, overwrite=overwrite, **parsed_args)

    def print_or_store(self, node, output_fname, fileformat, other_args={}, overwrite=False):
        """
        Depending on the parameters, either print the (single) output file on screen, or
        stores the file(s) on disk.

        :param node: the Data node to print or store on disk
        :param output_fname: The filename to store the main file. If empty or None, print
             instead
        :param fileformat: a string to pass to the _exportstring method
        :param other_args: a dictionary with additional kwargs to pass to _exportstring
        :param overwrite: if False, stops if any file already exists (when output_fname
             is not empty

        :note: this function calls directly sys.exit(1) when an error occurs (or e.g. if
            check_overwrite is True and a file already exists).
        """
        try:
            if output_fname:
                try:
                    node.export(
                        output_fname, fileformat=fileformat, overwrite=overwrite, **other_args)
                except OSError as e:
                    print >> sys.stderr, "verdi: ERROR while exporting file:"
                    print >> sys.stderr, e.message
                    sys.exit(1)
            else:
                filetext, extra_files = node._exportstring(
                    fileformat, main_file_name=output_fname, **other_args)
                if extra_files:
                    print >> sys.stderr, "This format requires to write more than one file."
                    print >> sys.stderr, "You need to pass the -o option to specify a file name."
                    sys.exit(1)
                else:
                    print filetext
        except TypeError as e:
            # This typically occurs for parameters that are passed down to the
            # methods in, e.g., BandsData, but they are not accepted
            print >> sys.stderr, "verdi: ERROR, probably a parameter is not supported by the specific format."
            print >> sys.stderr, "Error message: {}".format(e.message)
            raise

            sys.exit(1)


class Importable(object):
    """
    Provides shell completion for importable data nodes.

    .. note:: classes, inheriting Importable, MUST NOT contain attributes,
        starting with ``_import_``, which are not plugins for importing.
    """
    import_prefix = '_import_'
    import_parameters_postfix = '_parameters'

    def get_import_plugins(self):
        """
        Get the list of all implemented importers for data class.
        """
        method_names = dir(self)  # get list of class methods names
        valid_formats = [i[len(self.import_prefix):] for i in method_names
                         if i.startswith(self.import_prefix) and \
                         not i.endswith(self.import_parameters_postfix)]  # filter

        return {k: getattr(self, self.import_prefix + k) for k in valid_formats}

    def importfile(self, *args):
        import argparse, sys

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Import data object.')
        parser.add_argument('--file', type=str, default=None,
                            help="Path of the imported file. Reads from "
                                 "standard input if not specified.")

        default_format = None
        try:
            default_format = self._default_import_format
        except AttributeError:
            if len(self.get_import_plugins().keys()) == 1:
                default_format = self.get_import_plugins().keys()[0]
            else:
                default_format = None

        parser.add_argument('--format', '-f', type=str, default=default_format,
                            help="Type of the imported file.",
                            choices=sorted(self.get_import_plugins().keys()))

        # Augmenting the command line parameters with ones, that are used by
        # individual plugins
        for cmd in dir(self):
            if not cmd.startswith(self.import_prefix) or \
                    not cmd.endswith(self.import_parameters_postfix):
                continue
            getattr(self, cmd)(parser)

        args = list(args)
        parsed_args = vars(parser.parse_args(args))

        format = parsed_args.pop('format')
        filename = parsed_args.pop('file')

        if format is None:
            print >> sys.stderr, (
                "Default format is not defined, please specify.\n"
                "Valid formats are:")
            for i in self.get_import_plugins().keys():
                print >> sys.stderr, "  {}".format(i)
            sys.exit(1)

        if not filename:
            filename = "/dev/stdin"

        try:
            func = self.get_import_plugins()[format]
        except KeyError:
            print >> sys.stderr, "Not implemented; implemented plugins are:"
            print >> sys.stderr, "{}.".format(
                ",".join(self.get_import_plugins()))
            sys.exit(1)

        if not is_dbenv_loaded():
            load_dbenv()

        func(filename, **parsed_args)


class Depositable(object):
    """
    Provides shell completion for depositable data nodes.

    .. note:: classes, inheriting Depositable, MUST NOT contain
        attributes, starting with ``_deposit_``, which are not plugins for
        depositing.
    """
    deposit_prefix = '_deposit_'
    deposit_parameters_postfix = '_parameters'

    def get_deposit_plugins(self):
        """
        Get the list of all implemented deposition methods for data class.
        """
        method_names = dir(self)  # get list of class methods names
        valid_formats = [i[len(self.deposit_prefix):] for i in method_names
                         if i.startswith(self.deposit_prefix) and \
                         not i.endswith(self.deposit_parameters_postfix)]  # filter

        return {k: getattr(self, self.deposit_prefix + k) for k in valid_formats}

    def deposit(self, *args):
        """
        Deposit the data node to a given database.

        :param args: a namespace with parsed command line parameters.
        """
        # DEVELOPER NOTE: to add a new plugin, just add a _deposit_xxx() method.
        import argparse
        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Deposit data object.')
        parser.add_argument('data_id', type=int, default=None,
                            help="ID of the data object to be deposited.")

        default_database = None
        try:
            default_database = self._default_deposition_database
        except AttributeError:
            if len(self.get_deposit_plugins().keys()) == 1:
                default_database = self.get_deposit_plugins().keys()[0]
            else:
                default_database = None

        parser.add_argument('--database', '-d', type=str, default=default_database,
                            help="Label of the database for deposition.",
                            choices=self.get_deposit_plugins().keys())

        # Augmenting the command line parameters with ones, that are used by
        # individual plugins
        for cmd in dir(self):
            if not cmd.startswith(self.deposit_prefix) or \
                    not cmd.endswith(self.deposit_parameters_postfix):
                continue
            getattr(self, cmd)(parser)

        args = list(args)
        parsed_args = vars(parser.parse_args(args))

        database = parsed_args.pop('database')
        data_id = parsed_args.pop('data_id')

        # Removing the keys, whose values are None
        for key in parsed_args.keys():
            if parsed_args[key] is None:
                parsed_args.pop(key)

        if database is None:
            print >> sys.stderr, (
                "Default database is not defined, please specify.\n"
                "Valid databases are:")
            for i in self.get_deposit_plugins().keys():
                print >> sys.stderr, "  {}".format(i)
            sys.exit(1)

        try:
            func = self.get_deposit_plugins()[database]
        except KeyError:
            print >> sys.stderr, "Not implemented; implemented plugins are:"
            print >> sys.stderr, "{}.".format(
                ",".join(self.get_deposit_plugins()))
            sys.exit(1)

        if not is_dbenv_loaded():
            load_dbenv()

        n = load_node(data_id)

        try:
            if not isinstance(n, self.dataclass):
                print >> sys.stderr, ("Node {} is of class {} instead "
                                      "of {}".format(n, type(n), self.dataclass))
                sys.exit(1)
        except AttributeError:
            pass

        calc = func(n, **parsed_args)
        print calc


# Note: this class should not be exposed directly in the main module,
# otherwise it becomes a command of 'verdi'. Instead, we want it to be a
# subcommand of verdi data.
class _Upf(VerdiCommandWithSubcommands, Importable):
    """
    Setup and manage upf to be used

    This command allows to list and configure upf.
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        if not is_dbenv_loaded():
            load_dbenv()
        from aiida.orm.data.upf import UpfData

        self.dataclass = UpfData
        self.valid_subcommands = {
            'uploadfamily': (self.uploadfamily, self.complete_auto),
            'listfamilies': (self.listfamilies, self.complete_none),
            'import': (self.importfile, self.complete_none),
            'exportfamily': (self.exportfamily, self.complete_auto)
        }

    def uploadfamily(self, *args):
        """
        Upload a new pseudopotential family.

        Returns the numbers of files found and the number of nodes uploaded.

        Call without parameters to get some help.
        """
        import os.path

        if not len(args) == 3 and not len(args) == 4:
            print >> sys.stderr, ("After 'upf uploadfamily' there should be three "
                                  "arguments:")
            print >> sys.stderr, ("folder, group_name, group_description "
                                  "[OPTIONAL: --stop-if-existing]\n")
            sys.exit(1)

        folder = os.path.abspath(args[0])
        group_name = args[1]
        group_description = args[2]
        stop_if_existing = False

        if len(args) == 4:
            if args[3] == "--stop-if-existing":
                stop_if_existing = True
            else:
                print >> sys.stderr, 'Unknown directive: ' + args[3]
                sys.exit(1)

        if (not os.path.isdir(folder)):
            print >> sys.stderr, 'Cannot find directory: ' + folder
            sys.exit(1)

        import aiida.orm.data.upf as upf

        files_found, files_uploaded = upf.upload_upf_family(folder, group_name,
                                                            group_description, stop_if_existing)

        print "UPF files found: {}. New files uploaded: {}".format(files_found, files_uploaded)

    def listfamilies(self, *args):
        """
        Print on screen the list of upf families installed
        """
        # note that the following command requires that the upfdata has a
        # key called element. As such, it is not well separated.
        import argparse

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='List AiiDA upf families.')
        parser.add_argument('-e', '--element', nargs='+', type=str, default=None,
                            help="Filter the families only to those containing "
                                 "a pseudo for each of the specified elements")
        parser.add_argument('-d', '--with-description',
                            dest='with_description', action='store_true',
                            help="Show also the description for the UPF family")
        parser.set_defaults(with_description=False)

        args = list(args)
        parsed_args = parser.parse_args(args)

        from aiida.orm import DataFactory
        from aiida.orm.data.upf import UPFGROUP_TYPE

        UpfData = DataFactory('upf')
        from aiida.orm.querybuilder import QueryBuilder
        from aiida.orm.group import Group
        qb = QueryBuilder()
        qb.append(UpfData, tag='upfdata')
        if parsed_args.element is not None:
            qb.add_filter(UpfData, {'attributes.element': {'in': parsed_args.element}})
        qb.append(
            Group,
            group_of='upfdata', tag='group',
            project=["name", "description"],
            filters={"type": {'==': UPFGROUP_TYPE}}
        )

        qb.distinct()
        if qb.count() > 0:
            for res in qb.dict():
                group_name = res.get("group").get("name")
                group_desc = res.get("group").get("description")
                qb = QueryBuilder()
                qb.append(
                    Group,
                    tag='thisgroup',
                    filters={"name": {'like': group_name}}
                )
                qb.append(
                    UpfData,
                    project=["id"],
                    member_of='thisgroup'
                )

                if parsed_args.with_description:
                    description_string = ": {}".format(group_desc)
                else:
                    description_string = ""

                print "* {} [{} pseudos]{}".format(group_name, qb.count(),
                                                   description_string)

        else:
            print "No valid UPF pseudopotential family found."

    def exportfamily(self, *args):
        """
        Export a pseudopotential family into a folder.
        Call without parameters to get some help.
        """
        import os
        from aiida.common.exceptions import NotExistent
        from aiida.orm import DataFactory

        if not len(args) == 2:
            print >> sys.stderr, ("After 'upf export' there should be two "
                                  "arguments:")
            print >> sys.stderr, ("folder, upf_family_name\n")
            sys.exit(1)

        folder = os.path.abspath(args[0])
        group_name = args[1]

        UpfData = DataFactory('upf')
        try:
            group = UpfData.get_upf_group(group_name)
        except NotExistent:
            print >> sys.stderr, ("upf family {} not found".format(group_name))
            sys.exit(1)

        for u in group.nodes:
            dest_path = os.path.join(folder, u.filename)
            if not os.path.isfile(dest_path):
                with open(dest_path, 'w') as dest:
                    with u._get_folder_pathsubfolder.open(u.filename) as source:
                        dest.write(source.read())
            else:
                print >> sys.stdout, ("File {} is already present in the "
                                      "destination folder".format(u.filename))

    def _import_upf(self, filename, **kwargs):
        """
        Importer from UPF.
        """
        try:
            node, _ = self.dataclass.get_or_create(filename)
            print node
        except ValueError as e:
            print e


class _Bands(VerdiCommandWithSubcommands, Listable, Visualizable, Exportable):
    """
    Manipulation on the bands
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        if not is_dbenv_loaded():
            load_dbenv()
        from aiida.orm.data.array.bands import BandsData

        self.dataclass = BandsData
        self.valid_subcommands = {
            'show': (self.show, self.complete_none),
            'list': (self.list, self.complete_none),
            'export': (self.export, self.complete_none),
        }

    def query(self, args):
        """
        Perform the query and return information for the list.

        :param args: a namespace with parsed command line parameters.
        :return: table (list of lists) with information, describing nodes.
            Each row describes a single hit.
        """
        from aiida.backends.utils import QueryFactory
        if not is_dbenv_loaded():
            load_dbenv()

        q = QueryFactory()()
        return q.get_bands_and_parents_structure(args)

    def append_list_cmdline_arguments(self, parser):
        """
        Append additional command line parameters, that are later parsed and
        used in the query construction.

        :param parser: instance of argparse.ArgumentParser
        """
        parser.add_argument('-e', '--element', nargs='+', type=str, default=None,
                            help="Print all bandsdatas from structures "
                                 "containing desired elements")
        parser.add_argument('-eo', '--element-only', nargs='+', type=str, default=None,
                            help="Print all bandsdatas from structures "
                                 "containing only the selected elements")
        parser.add_argument('-f', '--formulamode', metavar='FORMULA_MODE',
                            type=str, default='hill',
                            help="Formula printing mode (hill, hill_compact,"
                                 " reduce, group, count, or count_compact)"
                                 " (if None, does not print the formula)",
                            action='store')
        parser.add_argument('-p', '--past-days', metavar='N',
                            help="Add a filter to show only bandsdatas created in the past N days",
                            type=int, action='store')
        parser.add_argument('-g', '--group-name', metavar='N', nargs="+", default=None,
                            help="add a filter to show only objects belonging to groups",
                            type=str, action='store')
        parser.add_argument('-G', '--group-pk', metavar='N', nargs="+", default=None,
                            help="add a filter to show only objects belonging to groups",
                            type=int, action='store')
        parser.add_argument('-A', '--all-users', action='store_true', default=False,
                            help="show groups for all users, rather than only for the"
                                 "current user")

    def append_export_cmdline_arguments(self, parser):
        """
        Additional command line arguments for the 'export' command

        :param parser: instance of argparse.ArgumentParser
        """
        from aiida.common.utils import Prettifier

        parser.add_argument('--prettify-format', type=str, default=None,
                            choices=Prettifier.get_prettifiers(),
                            help='The style of labels for the prettifier')
        parser.add_argument('--y-min-lim', type=float, default=None,
                            help='The minimum value for the y axis. Default: minimum of all bands')
        parser.add_argument('--y-max-lim', type=float, default=None,
                            help='The maximum value for the y axis. Default: maximum of all bands')

    def get_column_names(self):
        """
        Return the list with column names.

        :note: neither the number nor correspondence of column names and
            actual columns in the output from the query() are checked.
        """
        return ["ID", "formula", "ctime", "label"]

    def _export_agr(self, node, output_fname, overwrite, **kwargs):
        """
        Export a .agr file, to be visualized with the XMGrace plotting software.
        """
        self.print_or_store(node, output_fname, fileformat='agr', overwrite=overwrite,
                            other_args=kwargs)

    def _export_agr_batch(self, node, output_fname, overwrite, **kwargs):
        """
        Export a .agr batch file, to be visualized with the XMGrace plotting software.
        """
        self.print_or_store(node, output_fname, fileformat='agr_batch', overwrite=overwrite,
                            other_args=kwargs)

    def _export_gnuplot(self, node, output_fname, overwrite, **kwargs):
        """
        Export a Gnuplot file, together with the corresponding .dat file,
        to be visualized with the Gnuplot plotting software.

        Run with 'gnuplot -p filename' to see the plot (and keep the window with
        the plot open).
        """
        self.print_or_store(node, output_fname, fileformat='gnuplot', overwrite=overwrite,
                            other_args=kwargs)

    def _export_dat_multicolumn(self, node, output_fname, overwrite, **kwargs):
        """
        Export a .dat file with one line per kpoint, with multiple energy values
        on the same line separated by spaces.
        """
        self.print_or_store(node, output_fname, fileformat='dat_multicolumn', overwrite=overwrite,
                            other_args=kwargs)

    def _export_dat_blocks(self, node, output_fname, overwrite, **kwargs):
        """
        Export a .dat file with one line per datapoint (kpt, energy),
        with multiple bands separated in stanzas (i.e. having at least an empty
        newline inbetween).
        """
        self.print_or_store(node, output_fname, fileformat='dat_blocks', overwrite=overwrite,
                            other_args=kwargs)

    def _export_json(self, node, output_fname, overwrite, **kwargs):
        """
        Export a .dat file with one line per datapoint (kpt, energy),
        with multiple bands separated in stanzas (i.e. having at least an empty
        newline inbetween).
        """
        self.print_or_store(node, output_fname, fileformat='json', overwrite=overwrite,
                            other_args=kwargs)

    def _export_mpl_singlefile(self, node, output_fname, overwrite, **kwargs):
        """
        Export a .py file that would produce the plot using matplotlib
        when run with python (with data dumped within the same python file)
        """
        self.print_or_store(node, output_fname, fileformat='mpl_singlefile', overwrite=overwrite,
                            other_args=kwargs)

    def _export_mpl_withjson(self, node, output_fname, overwrite, **kwargs):
        """
        Export a .py file that would produce the plot using matplotlib
        when run with python (with data dumped in an external json filee)
        """
        self.print_or_store(node, output_fname, fileformat='mpl_withjson', overwrite=overwrite,
                            other_args=kwargs)

    def _export_mpl_png(self, node, output_fname, overwrite, **kwargs):
        """
        Export a .png file generated using matplotlib
        """
        if not output_fname:
            print >> sys.stderr, "To export to PNG please always specify the filename with the -o option"
            sys.exit(1)
        self.print_or_store(node, output_fname, fileformat='mpl_png', overwrite=overwrite,
                            other_args=kwargs)

    def _export_mpl_pdf(self, node, output_fname, overwrite, **kwargs):
        """
        Export a .pdf file generated using matplotlib
        """
        if not output_fname:
            print >> sys.stderr, "To export to PDF please always specify the filename with the -o option"
            sys.exit(1)
        self.print_or_store(node, output_fname, fileformat='mpl_pdf', overwrite=overwrite,
                            other_args=kwargs)

    def _show_xmgrace(self, exec_name, list_bands):
        """
        Plugin for show the bands with the XMGrace plotting software.
        """
        import tempfile, subprocess, numpy
        from aiida.orm.data.array.bands import max_num_agr_colors

        list_files = []
        current_band_number = 0
        for iband, bands in enumerate(list_bands):
            # extract number of bands
            nbnds = bands.get_bands().shape[1]
            text, _ = bands._exportstring('agr', setnumber_offset=current_band_number,
                                          color_number=numpy.mod(iband + 1, max_num_agr_colors))
            # write a tempfile
            f = tempfile.NamedTemporaryFile(suffix='.agr')
            f.write(text)
            f.flush()
            list_files.append(f)
            # update the number of bands already plotted
            current_band_number += nbnds

        try:
            subprocess.check_output([exec_name] + [f.name for f in list_files])
            _ = [f.close() for f in list_files]
        except subprocess.CalledProcessError:
            # The program died: just print a message
            print "Note: the call to {} ended with an error.".format(
                exec_name)
            _ = [f.close() for f in list_files]
        except OSError as e:
            _ = [f.close() for f in list_files]
            if e.errno == 2:
                print ("No executable '{}' found. Add to the path, "
                       "or try with an absolute path.".format(
                    exec_name))
                sys.exit(1)
            else:
                raise


class _Structure(VerdiCommandWithSubcommands,
                 Listable,
                 Visualizable,
                 Exportable,
                 Importable,
                 Depositable):
    """
    Visualize AiIDA structures
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        if not is_dbenv_loaded():
            load_dbenv()
        from aiida.orm.data.structure import StructureData

        super(_Structure, self).__init__()

        self.dataclass = StructureData
        self.valid_subcommands = {
            'show': (self.show, self.complete_none),
            'list': (self.list, self.complete_none),
            'export': (self.export, self.complete_none),
            'deposit': (self.deposit, self.complete_none),
            'import': (self.importfile, self.complete_none),
        }

    def query(self, args):
        """
        Perform the query
        """
        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.orm.querybuilder import QueryBuilder
        from aiida.orm.data.structure import StructureData
        from aiida.orm.implementation import Group
        from aiida.orm.data.structure import (get_formula, get_symbols_string)
        from aiida.orm.user import User

        qb = QueryBuilder()
        if args.all_users is False:
            user = self.backend.users.get_automatic_user()
            qb.append(User, tag="creator", filters={"email": user.email})
        else:
            qb.append(User, tag="creator")

        st_data_filters = {}
        self.query_past_days(st_data_filters, args)
        qb.append(StructureData, tag="struc", created_by="creator",
                  filters=st_data_filters,
                  project=["id", "label", "attributes.kinds",
                           "attributes.sites"])

        group_filters = {}
        self.query_group(group_filters, args)
        if group_filters:
            qb.append(Group, tag="group", filters=group_filters,
                      group_of="struc")

        struc_list_data = qb.distinct()

        entry_list = []
        if struc_list_data.count() > 0:
            for [id, label, akinds, asites] in struc_list_data.all():

                # If symbols are defined there is a filtering of the structures
                # based on the element
                # When QueryBuilder will support this (attribute)s filtering,
                # it will be pushed in the query.
                if args.element is not None:
                    all_symbols = [_["symbols"][0] for _ in akinds]
                    if not any([s in args.element for s in all_symbols]
                               ):
                        continue

                    if args.elementonly:
                        print "Not implemented elementonly search"
                        sys.exit(1)

                # We want only the StructureData that have attributes
                if akinds is None or asites is None:
                    continue

                symbol_dict = {}
                for k in akinds:
                    symbols = k['symbols']
                    weights = k['weights']
                    symbol_dict[k['name']] = get_symbols_string(symbols,
                                                                weights)

                try:
                    symbol_list = []
                    for s in asites:
                        symbol_list.append(symbol_dict[s['kind_name']])
                    formula = get_formula(symbol_list,
                                          mode=args.formulamode)
                # If for some reason there is no kind with the name
                # referenced by the site
                except KeyError:
                    formula = "<<UNKNOWN>>"
                entry_list.append([str(id), str(formula), label])

        return entry_list

    def append_list_cmdline_arguments(self, parser):
        parser.add_argument('-e', '--element', nargs='+', type=str, default=None,
                            help="Print all structures containing desired elements")
        parser.add_argument('-eo', '--elementonly', action='store_true',
                            help="If set, structures do not contain different "
                                 "elements (to be used with -e option)")
        parser.add_argument('-f', '--formulamode', metavar='FORMULA_MODE',
                            type=str, default='hill',
                            help="Formula printing mode (hill, hill_compact,"
                                 " reduce, group, count, or count_compact)"
                                 " (if None, does not print the formula)",
                            action='store')
        parser.add_argument('-p', '--past-days', metavar='N',
                            help="Add a filter to show only structures created in the past N days",
                            type=int, action='store')
        parser.add_argument('-g', '--group-name', metavar='N', nargs="+", default=None,
                            help="add a filter to show only objects belonging to groups",
                            type=str, action='store')
        parser.add_argument('-G', '--group-pk', metavar='N', nargs="+", default=None,
                            help="add a filter to show only objects belonging to groups",
                            type=int, action='store')
        parser.add_argument('-A', '--all-users', action='store_true', default=False,
                            help="show groups for all users, rather than only for the"
                                 "current user")

    def get_column_names(self):
        return ["ID", "formula", "label"]

    def _show_xcrysden(self, exec_name, structure_list):
        """
        Plugin for xcrysden
        """
        import tempfile, subprocess

        if len(structure_list) > 1:
            raise MultipleObjectsError("Visualization of multiple objects "
                                       "is not implemented")
        structure = structure_list[0]

        with tempfile.NamedTemporaryFile(suffix='.xsf') as f:
            f.write(structure._exportstring('xsf')[0])
            f.flush()

            try:
                subprocess.check_output([exec_name, '--xsf', f.name])
            except subprocess.CalledProcessError:
                # The program died: just print a message
                print "Note: the call to {} ended with an error.".format(
                    exec_name)
            except OSError as e:
                if e.errno == 2:
                    print ("No executable '{}' found. Add to the path, "
                           "or try with an absolute path.".format(
                        exec_name))
                    sys.exit(1)
                else:
                    raise

    def _show_ase(self, exec_name, structure_list):
        """
        Plugin to show the structure with the ASE visualizer
        """
        try:
            from ase.visualize import view
            for structure in structure_list:
                view(structure.get_ase())
        except ImportError:
            raise

    def _show_vesta(self, exec_name, structure_list):
        """
        Plugin for VESTA
        This VESTA plugin was added by Yue-Wen FANG and Abel Carreras
        at Kyoto University in the group of Prof. Isao Tanaka's lab

        """
        import tempfile, subprocess

        with tempfile.NamedTemporaryFile(suffix='.cif') as f:
            for structure in structure_list:
                f.write(structure._exportstring('cif')[0])
            f.flush()

            try:
                subprocess.check_output([exec_name, f.name])
            except subprocess.CalledProcessError:
                # The program died: just print a message
                print "Note: the call to {} ended with an error.".format(
                    exec_name)
            except OSError as e:
                if e.errno == 2:
                    print ("No executable '{}' found. Add to the path, "
                           "or try with an absolute path.".format(exec_name))
                    sys.exit(1)
                else:
                    raise

    def _show_vmd(self, exec_name, structure_list):
        """
        Plugin for vmd
        """
        import tempfile, subprocess

        if len(structure_list) > 1:
            raise MultipleObjectsError("Visualization of multiple objects "
                                       "is not implemented")
        structure = structure_list[0]

        with tempfile.NamedTemporaryFile(suffix='.xsf') as f:
            f.write(structure._exportstring('xsf')[0])
            f.flush()

            try:
                subprocess.check_output([exec_name, f.name])
            except subprocess.CalledProcessError:
                # The program died: just print a message
                print "Note: the call to {} ended with an error.".format(
                    exec_name)
            except OSError as e:
                if e.errno == 2:
                    print ("No executable '{}' found. Add to the path, "
                           "or try with an absolute path.".format(
                        exec_name))
                    sys.exit(1)
                else:
                    raise

    def _show_jmol(self, exec_name, structure_list):
        """
        Plugin for jmol
        """
        import tempfile, subprocess

        with tempfile.NamedTemporaryFile() as f:
            for structure in structure_list:
                f.write(structure._exportstring('cif')[0])
            f.flush()

            try:
                subprocess.check_output([exec_name, f.name])
            except subprocess.CalledProcessError:
                # The program died: just print a message
                print "Note: the call to {} ended with an error.".format(
                    exec_name)
            except OSError as e:
                if e.errno == 2:
                    print ("No executable '{}' found. Add to the path, "
                           "or try with an absolute path.".format(
                        exec_name))
                    sys.exit(1)
                else:
                    raise

    def _export_tcod(self, node, output_fname, overwrite, parameter_data=None, **kwargs):
        """
        Plugin for TCOD
        """

        parameters = None
        if parameter_data is not None:
            from aiida.orm import DataFactory
            ParameterData = DataFactory('parameter')
            parameters = load_node(parameter_data, parent_class=ParameterData)
        self.print_or_store(node, output_fname, fileformat='tcod', overwrite=overwrite,
                            other_args=kwargs)

    def _export_tcod_parameters(self, parser, **kwargs):
        """
        Command line parameters for TCOD
        """
        from aiida.tools.dbexporters.tcod import extend_with_cmdline_parameters
        extend_with_cmdline_parameters(parser, self.dataclass.__name__)

    def _export_xsf(self, node, output_fname, overwrite, **kwargs):
        """
        Exporter to XSF.
        """
        self.print_or_store(node, output_fname, fileformat='xsf', overwrite=overwrite,
                            other_args=kwargs)

    def _export_cif(self, node, output_fname, overwrite, **kwargs):
        """
        Exporter to CIF.
        """
        self.print_or_store(node, output_fname, fileformat='cif', overwrite=overwrite,
                            other_args=kwargs)

    def _export_xyz(self, node, output_fname, overwrite, **kwargs):
        """
        Exporter to XYZ.
        """
        self.print_or_store(node, output_fname, fileformat='xyz', overwrite=overwrite,
                            other_args=kwargs)

    def _import_xyz_parameters(self, parser):
        """
        Adding some functionality to the parser to deal with importing files
        """
        # In order to deal with structures that do not have a cell defined:
        # We can increase the size of the cell from the minimal cell
        # The minimal cell is the cell the just accomodates the structure given,
        # defined by the minimum and maximum of position in each dimension
        parser.add_argument('--vacuum-factor', type=float, default=1.0,
                            help='The factor by which the cell accomodating the structure should be increased, default: 1.0')
        # To that increased cell, we can also add a "safety margin"
        parser.add_argument('--vacuum-addition', type=float, default=10.0,
                            help='The distance to add to the unit cell after vacuum-factor was applied to expand in each dimension, default: 10.0')
        parser.add_argument('--pbc', type=int, nargs=3, default=[0, 0, 0],
                            help="""
                Set periodic boundary conditions for each lattice direction,
                0 for no periodicity, any other integer for periodicity""")
        parser.add_argument('--view', action='store_true', default=False, help='View resulting structure using ASE')
        parser.add_argument('--dont-store', action='store_true', default=False, help='Do not store the structure')

    def _import_xyz(self, filename, **kwargs):
        """
        Imports an XYZ-file.
        """
        from os.path import abspath
        vacuum_addition = kwargs.pop('vacuum_addition')
        vacuum_factor = kwargs.pop('vacuum_factor')
        pbc = [bool(i) for i in kwargs.pop('pbc')]
        dont_store = kwargs.pop('dont_store')
        view_in_ase = kwargs.pop('view')

        print 'importing XYZ-structure from: \n  {}'.format(abspath(filename))
        filepath = abspath(filename)
        with open(filepath) as f:
            xyz_txt = f.read()
        new_structure = self.dataclass()
        try:
            new_structure._parse_xyz(xyz_txt)
            new_structure._adjust_default_cell(vacuum_addition=vacuum_addition,
                                               vacuum_factor=vacuum_factor,
                                               pbc=pbc)

            if not dont_store:
                new_structure.store()
            if view_in_ase:
                from ase.visualize import view
                view(new_structure.get_ase())
            print  (
                '  Succesfully imported structure {}, '
                '(PK = {})'.format(new_structure.get_formula(), new_structure.pk)
            )

        except ValueError as e:
            print e
            sys.exit(1)

    def _import_pwi(self, filename, **kwargs):
        """
        Imports a structure from a quantumespresso input file.
        """
        from os.path import abspath
        try:
            from qe_tools.parsers.pwinputparser import PwInputFile
        except ImportError:
            print ("You have not installed the package qe-tools. \n"
                   "You can install it with: pip install qe-tools")
            sys.exit(1)

        dont_store = kwargs.pop('dont_store')
        view_in_ase = kwargs.pop('view')

        print 'importing structure from: \n  {}'.format(abspath(filename))
        filepath = abspath(filename)

        try:
            inputparser = PwInputFile(filepath)
            new_structure = inputparser.get_structuredata()

            if not dont_store:
                new_structure.store()
            if view_in_ase:
                from ase.visualize import view
                view(new_structure.get_ase())
            print  (
                '  Succesfully imported structure {}, '
                '(PK = {})'.format(new_structure.get_formula(), new_structure.pk)
            )

        except ValueError as e:
            print e
            sys.exit(1)

    def _import_ase(self, filename, **kwargs):
        """
        Imports a structure in a number of formats using the ASE routines.
        """
        from os.path import abspath
        from aiida.orm.data.structure import StructureData

        try:
            import ase.io
        except ImportError:
            print ("You have not installed the package ase. \n"
                   "You can install it with: pip install ase")
            sys.exit(1)

        dont_store = kwargs.pop('dont_store')
        view_in_ase = kwargs.pop('view')

        print 'importing structure from: \n  {}'.format(abspath(filename))
        filepath = abspath(filename)

        try:
            asecell = ase.io.read(filepath)
            new_structure = StructureData(ase=asecell)

            if not dont_store:
                new_structure.store()
            if view_in_ase:
                from ase.visualize import view
                view(new_structure.get_ase())
            print  (
                '  Succesfully imported structure {}, '
                '(PK = {})'.format(new_structure.get_formula(), new_structure.pk)
            )

        except ValueError as e:
            print e
            sys.exit(1)

    def _deposit_tcod(self, node, parameter_data=None, **kwargs):
        """
        Deposition plugin for TCOD.
        """
        from aiida.tools.dbexporters.tcod import deposit

        parameters = None
        if parameter_data is not None:
            from aiida.orm import DataFactory
            ParameterData = DataFactory('parameter')
            parameters = load_node(parameter_data, parent_class=ParameterData)
        return deposit(node, parameters=parameters, **kwargs)

    def _deposit_tcod_parameters(self, parser, **kwargs):
        """
        Command line parameters deposition plugin for TCOD.
        """
        from aiida.tools.dbexporters.tcod import (deposition_cmdline_parameters,
                                                  extend_with_cmdline_parameters)
        deposition_cmdline_parameters(parser, self.dataclass.__name__)
        extend_with_cmdline_parameters(parser, self.dataclass.__name__)


class _Cif(VerdiCommandWithSubcommands,
           Listable, Visualizable, Exportable, Importable, Depositable):
    """
    Visualize CIF structures
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        if not is_dbenv_loaded():
            load_dbenv()
        from aiida.orm.data.cif import CifData
        super(_Cif, self).__init__()

        self.dataclass = CifData
        self.valid_subcommands = {
            'show': (self.show, self.complete_none),
            'list': (self.list, self.complete_none),
            'export': (self.export, self.complete_none),
            'import': (self.importfile, self.complete_none),
            'deposit': (self.deposit, self.complete_none),
        }

    def _show_jmol(self, exec_name, structure_list):
        """
        Plugin for jmol
        """
        import tempfile, subprocess

        with tempfile.NamedTemporaryFile() as f:
            for structure in structure_list:
                f.write(structure._exportstring('cif')[0])
            f.flush()

            try:
                subprocess.check_output([exec_name, f.name])
            except subprocess.CalledProcessError:
                # The program died: just print a message
                print "Note: the call to {} ended with an error.".format(
                    exec_name)
            except OSError as e:
                if e.errno == 2:
                    print ("No executable '{}' found. Add to the path, "
                           "or try with an absolute path.".format(
                        exec_name))
                    sys.exit(1)
                else:
                    raise

    def _show_vesta(self, exec_name, structure_list):
        """
        Plugin for VESTA, added by Yue-Wen FANG and Abel Carreras
        at Kyoto University in the group of Prof. Isao Tanaka's lab
        """
        import tempfile, subprocess

        with tempfile.NamedTemporaryFile(suffix='.cif') as f:
            for structure in structure_list:
                f.write(structure._exportstring('cif')[0])
            f.flush()

            try:
                subprocess.check_output([exec_name, f.name])
            except subprocess.CalledProcessError:
                # The program died: just print a message
                print "Note: the call to {} ended with an error.".format(exec_name)
            except OSError as e:
                if e.errno == 2:
                    print ("No executable '{}' found. Add to the path, "
                           "or try with an absolute path.".format(exec_name))
                    sys.exit(1)
                else:
                    raise

    def query(self, args):
        """
        Perform the query and return information for the list.

        :param args: a namespace with parsed command line parameters.
        :return: table (list of lists) with information, describing nodes.
            Each row describes a single hit.
        """
        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.orm.querybuilder import QueryBuilder
        from aiida.orm.implementation import Group
        from aiida.orm.user import User

        qb = QueryBuilder()
        if args.all_users is False:
            user = self.backend.users.get_automatic_user()
            qb.append(User, tag="creator", filters={"email": user.email})
        else:
            qb.append(User, tag="creator")

        st_data_filters = {}
        self.query_past_days(st_data_filters, args)
        qb.append(self.dataclass, tag="struc", created_by="creator",
                  filters=st_data_filters,
                  project=["*"])

        group_filters = {}
        self.query_group(group_filters, args)
        if group_filters:
            qb.append(Group, tag="group", filters=group_filters,
                      group_of="struc")

        qb.order_by({self.dataclass: {'ctime': 'asc'}})
        res = qb.distinct()

        entry_list = []
        if res.count() > 0:
            for [obj] in res.iterall():
                formulae = '?'
                try:
                    formulae = ",".join(obj.get_attr('formulae'))
                except AttributeError:
                    pass
                except TypeError:
                    pass
                source_uri = '?'
                try:
                    source_uri = obj.get_attr('source')['uri']
                except AttributeError:
                    pass
                except KeyError:
                    pass
                entry_list.append([str(obj.pk), formulae, source_uri])
        return entry_list

    def get_column_names(self):
        """
        Return the list with column names.

        :note: neither the number nor correspondence of column names and
            actual columns in the output from the query() are checked.
        """
        return ["ID", "formulae", "source_uri"]

    def _export_cif(self, node, output_fname, overwrite, **kwargs):
        """
        Exporter to CIF.
        """
        self.print_or_store(node, output_fname, fileformat='cif', overwrite=overwrite,
                            other_args=kwargs)

    def _export_tcod(self, node, output_fname, overwrite, parameter_data=None, **kwargs):
        """
        Plugin for TCOD
        """
        parameters = None
        if parameter_data is not None:
            from aiida.orm import DataFactory
            ParameterData = DataFactory('parameter')
            parameters = load_node(parameter_data, parent_class=ParameterData)
        self.print_or_store(node, output_fname, fileformat='tcod', overwrite=overwrite,
                            other_args=kwargs)

    def _export_tcod_parameters(self, parser, **kwargs):
        """
        Command line parameters for TCOD
        """
        from aiida.tools.dbexporters.tcod import extend_with_cmdline_parameters
        extend_with_cmdline_parameters(parser, self.dataclass.__name__)

    def _import_cif(self, filename, **kwargs):
        """
        Importer from CIF.
        """
        import os.path

        try:
            node, _ = self.dataclass.get_or_create(os.path.abspath(filename))
            print node
        except ValueError as e:
            print e

    def _deposit_tcod(self, node, parameter_data=None, **kwargs):
        """
        Deposition plugin for TCOD.
        """
        from aiida.tools.dbexporters.tcod import deposit

        parameters = None
        if parameter_data is not None:
            from aiida.orm import DataFactory
            ParameterData = DataFactory('parameter')
            parameters = load_node(parameter_data, parent_class=ParameterData)
        return deposit(node, parameters=parameters, **kwargs)

    def _deposit_tcod_parameters(self, parser, **kwargs):
        """
        Command line parameters deposition plugin for TCOD.
        """
        from aiida.tools.dbexporters.tcod import (deposition_cmdline_parameters,
                                                  extend_with_cmdline_parameters)
        deposition_cmdline_parameters(parser, self.dataclass.__name__)
        extend_with_cmdline_parameters(parser, self.dataclass.__name__)


class _Trajectory(VerdiCommandWithSubcommands,
                  Listable, Visualizable, Exportable, Depositable):
    """
    View and manipulate TrajectoryData instances.
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        if not is_dbenv_loaded():
            load_dbenv()
        from aiida.orm.data.array.trajectory import TrajectoryData

        self.dataclass = TrajectoryData
        self.valid_subcommands = {
            'show': (self.show, self.complete_none),
            'list': (self.list, self.complete_none),
            'export': (self.export, self.complete_none),
            'deposit': (self.deposit, self.complete_none),
        }

    def _show_jmol(self, exec_name, trajectory_list, **kwargs):
        """
        Plugin for jmol
        """
        import tempfile, subprocess

        with tempfile.NamedTemporaryFile() as f:
            for trajectory in trajectory_list:
                f.write(trajectory._exportstring('cif', **kwargs)[0])
            f.flush()

            try:
                subprocess.check_output([exec_name, f.name])
            except subprocess.CalledProcessError:
                # The program died: just print a message
                print "Note: the call to {} ended with an error.".format(
                    exec_name)
            except OSError as e:
                if e.errno == 2:
                    print ("No executable '{}' found. Add to the path, "
                           "or try with an absolute path.".format(
                        exec_name))
                    sys.exit(1)
                else:
                    raise

    def _show_jmol_parameters(self, parser):
        """
        Describe command line parameters.
        """
        parser.add_argument('--step',
                            help="ID of the trajectory step. If none is "
                                 "supplied, all steps are exported.",
                            type=int, action='store')

    def _show_xcrysden(self, exec_name, trajectory_list, **kwargs):
        """
        Plugin for xcrysden
        """
        import tempfile, subprocess

        if len(trajectory_list) > 1:
            raise MultipleObjectsError("Visualization of multiple trajectories "
                                       "is not implemented")
        trajectory = trajectory_list[0]

        with tempfile.NamedTemporaryFile(suffix='.xsf') as f:
            f.write(trajectory._exportstring('xsf', **kwargs)[0])
            f.flush()

            try:
                subprocess.check_output([exec_name, '--xsf', f.name])
            except subprocess.CalledProcessError:
                # The program died: just print a message
                print "Note: the call to {} ended with an error.".format(
                    exec_name)
            except OSError as e:
                if e.errno == 2:
                    print ("No executable '{}' found. Add to the path, "
                           "or try with an absolute path.".format(
                        exec_name))
                    sys.exit(1)
                else:
                    raise

    def _show_mpl_pos_parameters(self, parser):
        """
        Describe command line parameters for _show_pos
        """
        parser.add_argument('-s', '--stepsize',
                            type=int,
                            help=''
                                 'The stepsize for the trajectory, set it higher to reduce '
                                 'number of points',
                            default=1
                            )
        parser.add_argument('--mintime',
                            type=int, default=None,
                            help='The time to plot from'
                            )
        parser.add_argument('--maxtime',
                            type=int, default=None,
                            help='The time to plot to'
                            )
        parser.add_argument('-e', '--elements',
                            type=str, nargs='+',
                            help='Show only atoms of that species'
                            )
        parser.add_argument('-i', '--indices',
                            type=int, nargs='+',
                            help='Show only these indices'
                            )
        parser.add_argument('--dont-block',
                            action='store_true',
                            help="Don't block interpreter when showing plot"
                            )

    def _show_mpl_heatmap_parameters(self, parser):
        """
        Describe command line parameters for _show_mpl_heatmap
        """
        parser.add_argument('-c', '--contours',
                            type=float, nargs='+',
                            help='Isovalues to plot'
                            )
        parser.add_argument('--sampling-stepsize',
                            type=int,
                            help='Sample positions in plot every sampling_stepsize timestep'
                            )

    def _show_mpl_pos(self, exec_name, trajectory_list, **kwargs):
        """
        Produces a matplotlib plot of the trajectory
        """
        for t in trajectory_list:
            t.show_mpl_pos(**kwargs)

    def _show_mpl_heatmap(self, exec_name, trajectory_list, **kwargs):
        """
        Produces a matplotlib plot of the trajectory
        """
        for t in trajectory_list:
            t.show_mpl_heatmap(**kwargs)

    def _export_xsf(self, node, output_fname, overwrite, **kwargs):
        """
        Exporter to XSF.
        """
        self.print_or_store(node, output_fname, fileformat='xsf', overwrite=overwrite,
                            other_args=kwargs)

    def _export_tcod(self, node, output_fname, overwrite, parameter_data=None, **kwargs):
        """
        Plugin for TCOD
        """

        parameters = None
        if parameter_data is not None:
            from aiida.orm import DataFactory
            ParameterData = DataFactory('parameter')
            parameters = load_node(parameter_data, parent_class=ParameterData)
        self.print_or_store(node, output_fname, fileformat='tcod', overwrite=overwrite,
                            other_args=kwargs)

    def _export_tcod_parameters(self, parser, **kwargs):
        """
        Command line parameters for TCOD
        """
        from aiida.tools.dbexporters.tcod import extend_with_cmdline_parameters
        extend_with_cmdline_parameters(parser, self.dataclass.__name__)

    def _export_cif(self, node, output_fname, overwrite, **kwargs):
        """
        Exporter to CIF.
        """
        self.print_or_store(node, output_fname, fileformat='cif', overwrite=overwrite,
                            other_args=kwargs)

    def _export_cif_parameters(self, parser, **kwargs):
        """
        Describe command line parameters.
        """
        parser.add_argument('--step', dest='trajectory_index',
                            help="ID of the trajectory step. If none is "
                                 "supplied, all steps are exported.",
                            type=int, action='store')

    def _deposit_tcod(self, node, parameter_data=None, **kwargs):
        """
        Deposition plugin for TCOD.
        """
        from aiida.tools.dbexporters.tcod import deposit

        parameters = None
        if parameter_data is not None:
            from aiida.orm import DataFactory
            ParameterData = DataFactory('parameter')
            parameters = load_node(parameter_data, parent_class=ParameterData)
        return deposit(node, parameters=parameters, **kwargs)

    def _deposit_tcod_parameters(self, parser, **kwargs):
        """
        Command line parameters deposition plugin for TCOD.
        """
        from aiida.tools.dbexporters.tcod import (deposition_cmdline_parameters,
                                                  extend_with_cmdline_parameters)
        deposition_cmdline_parameters(parser, self.dataclass.__name__)
        extend_with_cmdline_parameters(parser, self.dataclass.__name__)
        self._export_cif_parameters(parser)


class _Parameter(VerdiCommandWithSubcommands, Visualizable):
    """
    View and manipulate Parameter data classes.
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        if not is_dbenv_loaded():
            load_dbenv()
        from aiida.orm.data.parameter import ParameterData

        self.dataclass = ParameterData
        self._default_show_format = 'json_date'
        self.valid_subcommands = {
            'show': (self.show, self.complete_none),
        }

    def _show_json_date(self, exec_name, node_list):
        """
        Show contents of ParameterData nodes.
        """
        from aiida.cmdline import print_dictionary

        for node in node_list:
            the_dict = node.get_dict()
            print_dictionary(the_dict, 'json+date')


class _Array(VerdiCommandWithSubcommands, Visualizable):
    """
    View and manipulate Array data classes.
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.orm.data.array import ArrayData

        self.dataclass = ArrayData
        self._default_show_format = 'json_date'
        self.valid_subcommands = {
            'show': (self.show, self.complete_none),
        }

    def _show_json_date(self, exec_name, node_list):
        """
        Show contents of ArrayData nodes.
        """
        from aiida.cmdline import print_dictionary

        for node in node_list:
            the_dict = {}
            for arrayname in node.arraynames():
                the_dict[arrayname] = node.get_array(arrayname).tolist()
            print_dictionary(the_dict, 'json+date')


class _Remote(VerdiCommandWithSubcommands):
    """
    Manage RemoteData objects
    """

    def __init__(self):
        self.valid_subcommands = {
            'ls': (self.do_listdir, self.complete_none),
            'cat': (self.do_cat, self.complete_none),
            'show': (self.do_show, self.complete_none),
        }

    def do_listdir(self, *args):
        """
        List directory content on remote RemoteData objects.
        """
        import argparse
        import datetime
        from aiida.backends.utils import load_dbenv, is_dbenv_loaded
        from aiida.common.utils import get_mode_string

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='List directory content on remote RemoteData objects.')

        parser.add_argument('-l', '--long', action='store_true',
                            help="Display also file metadata")
        parser.add_argument('pk', type=int, help="PK of the node")
        parser.add_argument('path', nargs='?', default='.', help="The folder to list")

        args = list(args)
        parsed_args = parser.parse_args(args)

        if not is_dbenv_loaded():
            load_dbenv()

        try:
            n = load_node(parsed_args.pk)
        except Exception as e:
            click.echo(e.message, err=True)
            sys.exit(1)
        try:
            content = n.listdir_withattributes(path=parsed_args.path)
        except (IOError, OSError) as e:
            click.echo("Unable to access the remote folder or file, check if it exists.", err=True)
            click.echo("Original error: {}".format(str(e)), err=True)
            sys.exit(1)
        for metadata in content:
            if parsed_args.long:
                mtime = datetime.datetime.fromtimestamp(
                    metadata['attributes'].st_mtime)
                pre_line = '{} {:10}  {}  '.format(
                    get_mode_string(metadata['attributes'].st_mode),
                    metadata['attributes'].st_size,
                    mtime.strftime("%d %b %Y %H:%M")
                )
                click.echo(pre_line, nl=False)
            if metadata['isdir']:
                click.echo(click.style(metadata['name'], fg='blue'))
            else:
                click.echo(metadata['name'])

    def do_cat(self, *args):
        """
        Show the content of remote files in RemoteData objects.
        """
        # Note: the implementation is not very efficient: if first downloads the full file on a file on the disk,
        # then prints it and finally deletes the file.
        # TODO: change it to open the file and stream it; it requires to add an openfile() method to the transport
        import argparse
        import datetime
        from aiida.backends.utils import load_dbenv, is_dbenv_loaded
        import tempfile
        import os

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Show the content of remote files in RemoteData objects.')

        parser.add_argument('pk', type=int, help="PK of the node")
        parser.add_argument('path', type=str, help="The (relative) path to the file to show")

        args = list(args)
        parsed_args = parser.parse_args(args)

        if not is_dbenv_loaded():
            load_dbenv()

        try:
            n = load_node(parsed_args.pk)
        except Exception as e:
            click.echo(e.message, err=True)
            sys.exit(1)

        try:
            with tempfile.NamedTemporaryFile(delete=False) as f:
                f.close()
                n.getfile(parsed_args.path, f.name)
                with open(f.name) as fobj:
                    sys.stdout.write(fobj.read())
        except IOError as e:
            click.echo("ERROR {}: {}".format(e.errno, str(e)), err=True)
            sys.exit(1)

        try:
            os.remove(f.name)
        except OSError:
            # If you cannot delete, ignore (maybe I didn't manage to create it in the first place
            pass

    def do_show(self, *args):
        """
        Show information on a RemoteData object.
        """
        import argparse
        import datetime
        from aiida.backends.utils import load_dbenv, is_dbenv_loaded
        import tempfile
        import os

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Show information on a RemoteData object.')

        parser.add_argument('pk', type=int, help="PK of the node")

        args = list(args)
        parsed_args = parser.parse_args(args)

        if not is_dbenv_loaded():
            load_dbenv()

        try:
            n = load_node(parsed_args.pk)
        except Exception as e:
            click.echo(e.message, err=True)
            sys.exit(1)

        click.echo("- Remote computer name:")
        click.echo("  {}".format(n.get_computer_name()))
        click.echo("- Remote folder full path:")
        click.echo("  {}".format(n.get_remote_path()))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.orm.data import to_aiida_type, BaseType


def _left_operator(func):
    def inner(self, other):
        l = self.value
        if isinstance(other, NumericType):
            r = other.value
        else:
            r = other
        return to_aiida_type(func(l, r))
    return inner


def _right_operator(func):
    def inner(self, other):
        assert not isinstance(other, NumericType)
        return to_aiida_type(func(self.value, other))
    return inner


class NumericType(BaseType):
    """
    Specific subclass of :py:class:`aiida.orm.data.BaseType` to store numbers,
    overloading common operators (``+``, ``*``, ...)
    """
    @_left_operator
    def __add__(self, other):
        return self + other

    @_right_operator
    def __radd__(self, other):
        return other + self

    @_left_operator
    def __sub__(self, other):
        return self - other

    @_right_operator
    def __rsub__(self, other):
        return other - self

    @_left_operator
    def __mul__(self, other):
        return self * other

    @_right_operator
    def __rmul__(self, other):
        return other * self

    @_left_operator
    def __pow__(self, power):
        return self ** power

    @_left_operator
    def __lt__(self, other):
        return self < other

    @_left_operator
    def __le__(self, other):
        return self <= other

    @_left_operator
    def __gt__(self, other):
        return self > other

    @_left_operator
    def __ge__(self, other):
        return self >= other

    @_left_operator
    def __mod__(self, other):
        return self % other

    @_right_operator
    def __rmod__(self, other):
        return other % self

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Top level functions that can be used to launch a Process."""
from . import processes
from . import runners

__all__ = ['run', 'run_get_pid', 'run_get_node', 'submit']


def submit(process, **inputs):
    """
    Submit the process with the supplied inputs to the daemon runner immediately returning control to
    the interpreter. The return value will be the calculation node of the submitted process

    :param process: the process class to submit
    :param inputs: the inputs to be passed to the process
    :return: the calculation node of the process
    """
    runner = runners.new_runner(rmq_submit=True)
    return runner.submit(process, **inputs)


def run(process, *args, **inputs):
    """
    Run the process with the supplied inputs in a local runner that will block until the process is completed.
    The return value will be the results of the completed process

    :param process: the process class or workfunction to run
    :param inputs: the inputs to be passed to the process
    :return: the outputs of the process
    """
    if isinstance(process, processes.Process):
        runner = process.runner
    else:
        runner = runners.get_runner()
    return runner.run(process, *args, **inputs)


def run_get_node(process, *args, **inputs):
    """
    Run the process with the supplied inputs in a local runner that will block until the process is completed.
    The return value will be the results of the completed process

    :param process: the process class or workfunction to run
    :param inputs: the inputs to be passed to the process
    :return: tuple of the outputs of the process and the calculation node
    """
    if isinstance(process, processes.Process):
        runner = process.runner
    else:
        runner = runners.get_runner()
    return runner.run_get_node(process, *args, **inputs)


# Allow user to also use run.get_node as a convenience (one less import)
run.get_node = run_get_node


def run_get_pid(process, *args, **inputs):
    """
    Run the process with the supplied inputs in a local runner that will block until the process is completed.
    The return value will be the results of the completed process

    :param process: the process class or workfunction to run
    :param inputs: the inputs to be passed to the process
    :return: tuple of the outputs of the process and process pid
    """
    if isinstance(process, processes.Process):
        runner = process.runner
    else:
        runner = runners.get_runner()
    return runner.run_get_pid(process, *args, **inputs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida.common import aiidalogger
from aiida.common.datastructures import wf_states, wf_exit_call, wf_default_call


logger = aiidalogger.getChild('workflowmanager')


def execute_steps():
    """
    This method loops on the RUNNING workflows and handled the execution of the
    steps until each workflow reaches an end (or gets stopped for errors).

    In the loop for each RUNNING workflow the method loops also in each of its 
    RUNNING steps, testing if all the calculation and subworkflows attached to 
    the step are FINISHED. In this case the step is set as FINISHED and the 
    workflow is advanced to the step's next method present in the db with 
    ``advance_workflow``, otherwise if any step's JobCalculation is found in 
    NEW state the method will submit. If none of the previous conditions apply 
    the step is flagged as ERROR and cannot proceed anymore, blocking the future
    execution of the step and, connected, the workflow.

    Finally, for each workflow the method tests if there are INITIALIZED steps 
    to be launched, and in case reloads the workflow and execute the specific 
    those steps. In case or error the step is flagged in ERROR state and the 
    stack is reported in the workflow report.
    """

    from aiida.orm import JobCalculation
    from aiida.orm.implementation import get_all_running_steps
 
    logger.debug("Querying the worflow DB")
    
    running_steps = get_all_running_steps()

    for s in running_steps:
        if s.parent.state == wf_states.FINISHED:
            s.set_state(wf_states.FINISHED)
            continue

        w = s.parent.get_aiida_class()

        logger.info("[{0}] Found active step: {1}".format(w.pk, s.name))

        s_calcs_new = [c.pk for c in s.get_calculations() if c._is_new()]
        s_calcs_finished = [c.pk for c in s.get_calculations() if c.is_finished_ok]
        s_calcs_failed = [c.pk for c in s.get_calculations() if c.is_failed]
        s_calcs_num = len(s.get_calculations())

        s_sub_wf_finished = [sw.pk for sw in s.get_sub_workflows() if sw.has_finished_ok()]
        s_sub_wf_failed = [sw.pk for sw in s.get_sub_workflows() if sw.has_failed()]
        s_sub_wf_num = len(s.get_sub_workflows())

        if (s_calcs_num == (len(s_calcs_finished) + len(s_calcs_failed)) and
            s_sub_wf_num == (len(s_sub_wf_finished) + len(s_sub_wf_failed))):

            logger.info("[{0}] Step: {1} ready to move".format(w.pk, s.name))

            s.set_state(wf_states.FINISHED)
            
            advance_workflow(w, s)

        elif len(s_calcs_new) > 0:

            for pk in s_calcs_new:

                obj_calc = JobCalculation.get_subclass_from_pk(pk=pk)
                try:
                    obj_calc.submit()
                    logger.info("[{0}] Step: {1} launched calculation {2}".format(w.pk, s.name, pk))
                except:
                    logger.error("[{0}] Step: {1} cannot launch calculation {2}".format(w.pk, s.name, pk))


def advance_workflow(w, step):
    """
    The method tries to advance a step running its next method and handling 
    possible errors.

    If the method to advance is the Workflow ``exit`` method and there are no 
    more steps RUNNING or in ERROR state then the workflow is set to FINISHED, 
    otherwise an error is added to the report and the Workflow is flagged as 
    ERROR.

    If the method is the ``wf_default_call`` this means the step had no next, 
    and possibly is part of a branching. In this case the Workflow is not 
    advanced but the method returns True to let the other steps kick in.

    Finally the methos tries to load the Workflow and execute the selected step,
    reporting the errors and the stack trace in the report in case of problems. 
    Is no errors are reported the method returns True, in all the other cases 
    the Workflow is set to ERROR state and the method returns False.

    :param w: Workflow object to advance
    :param step: DbWorkflowStep to execute
    :return: True if the step has been executed, False otherwise
    """
    if step.nextcall == wf_exit_call:
        logger.info("[{0}] Step: {1} has an exit call".format(w.pk, step.name))
        if len(w.get_steps(wf_states.RUNNING)) == 0 and len(w.get_steps(wf_states.ERROR)) == 0:
            logger.info("[{0}] Step: {1} is really finished, going out".format(w.pk, step.name))
            w.set_state(wf_states.FINISHED)
            return True
        else:
            logger.error("[{0}] Step: {1} is NOT finished, stopping workflow "
                         "with error".format(w.pk, step.name))
            w.append_to_report("""Step: {0} is NOT finished, some calculations or workflows
            are still running and there is a next call, stopping workflow with error""".format(step.name))
            w.set_state(wf_states.ERROR)

            return False
    elif step.nextcall == wf_default_call:
        logger.info("[{0}] Step: {1} is not finished and has no next call, waiting "
                    "for other methods to kick.".format(w.pk, step.name))
        w.append_to_report("Step: {0} is not finished and has no "
                                      "next call, waiting for other methods "
                                      "to kick.".format(step.name))
        return True

    elif not step.nextcall is None:

        logger.info("[{0}] In advance_workflow the step {1} goes to nextcall {2}"
                    "".format(w.pk, step.name, step.nextcall))

        try:
            #w = Workflow.get_subclass_from_pk(w_superclass.pk)
            getattr(w, step.nextcall)()
            return True

        except Exception:
            import traceback

            w.append_to_report("ERROR ! This workflow got an error in the {0} "
                               "method, we report down the stack trace".format(
                                   step.nextcall))
            w.append_to_report("full traceback: {0}".format(traceback.format_exc()))

            w.get_step(step.nextcall).set_state(wf_states.ERROR)
            w.set_state(wf_states.ERROR)

            return False
    else:

        logger.error("[{0}] Step: {1} ERROR, no nextcall".format(w.pk,
                                                                 step.name))
        w.append_to_report("Step: {0} ERROR, no nextcall".format(step.name))
        w.set_state(wf_states.ERROR)

        return False





# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import threading

from plumpy.utils import AttributesFrozendict
from aiida import work
from aiida.backends.testbase import AiidaTestCase
from aiida.common.lang import override
from aiida.orm import load_node
from aiida.orm.data.int import Int
from aiida.orm.data.frozendict import FrozenDict
from aiida.orm.data.parameter import ParameterData
from aiida.work import test_utils, Process


class NameSpacedProcess(work.Process):

    @classmethod
    def define(cls, spec):
        super(NameSpacedProcess, cls).define(spec)
        spec.input('some.name.space.a', valid_type=Int)


class TestProcessNamespace(AiidaTestCase):

    def setUp(self):
        super(TestProcessNamespace, self).setUp()
        self.assertIsNone(Process.current())

    def tearDown(self):
        super(TestProcessNamespace, self).tearDown()
        self.assertIsNone(Process.current())

    def test_namespaced_process(self):
        """
        Test that inputs in nested namespaces are properly validated and the link labels
        are properly formatted by connecting the namespaces with underscores
        """
        proc = NameSpacedProcess(inputs={'some': {'name': {'space': {'a': Int(5)}}}})

        # Test that the namespaced inputs are AttributesFrozenDicts
        self.assertIsInstance(proc.inputs, AttributesFrozendict)
        self.assertIsInstance(proc.inputs.some, AttributesFrozendict)
        self.assertIsInstance(proc.inputs.some.name, AttributesFrozendict)
        self.assertIsInstance(proc.inputs.some.name.space, AttributesFrozendict)

        # Test that the input node is in the inputs of the process
        input_node = proc.inputs.some.name.space.a
        self.assertTrue(isinstance(input_node, Int))
        self.assertEquals(input_node.value, 5)

        # Check that the link of the WorkCalculation node has the correct link name
        self.assertTrue('some_name_space_a' in proc.calc.get_inputs_dict())
        self.assertEquals(proc.calc.get_inputs_dict()['some_name_space_a'], 5)


class ProcessStackTest(work.Process):
    @override
    def _run(self):
        pass

    @override
    def on_create(self):
        super(ProcessStackTest, self).on_create()
        self._thread_id = threading.current_thread().ident

    @override
    def on_stop(self):
        # The therad must match the one used in on_create because process
        # stack is using thread local storage to keep track of who called who
        super(ProcessStackTest, self).on_stop()
        assert self._thread_id is threading.current_thread().ident


class TestProcess(AiidaTestCase):

    def setUp(self):
        super(TestProcess, self).setUp()
        work.runners.set_runner(None)
        self.assertIsNone(Process.current())

    def tearDown(self):
        super(TestProcess, self).tearDown()
        self.assertIsNone(Process.current())

    def test_process_stack(self):
        work.launch.run(ProcessStackTest)

    def test_inputs(self):
        with self.assertRaises(TypeError):
            work.launch.run(test_utils.BadOutput)

    def test_input_link_creation(self):
        dummy_inputs = ["1", "2", "3", "4"]

        inputs = {l: Int(l) for l in dummy_inputs}
        inputs['store_provenance'] = True
        p = test_utils.DummyProcess(inputs)

        for label, value in p._calc.get_inputs_dict().iteritems():
            self.assertTrue(label in inputs)
            self.assertEqual(int(label), int(value.value))
            dummy_inputs.remove(label)

        # Make sure there are no other inputs
        self.assertFalse(dummy_inputs)

    def test_none_input(self):
        # Check that if we pass no input the process runs fine
        work.launch.run(test_utils.DummyProcess)

    def test_seal(self):
        pid = work.launch.run_get_pid(test_utils.DummyProcess).pid
        self.assertTrue(load_node(pk=pid).is_sealed)

    def test_description(self):
        dp = test_utils.DummyProcess(inputs={'description': "Rockin' process"})
        self.assertEquals(dp.calc.description, "Rockin' process")

        with self.assertRaises(ValueError):
            test_utils.DummyProcess(inputs={'description': 5})

    def test_label(self):
        dp = test_utils.DummyProcess(inputs={'label': 'My label'})
        self.assertEquals(dp.calc.label, 'My label')

        with self.assertRaises(ValueError):
            test_utils.DummyProcess(inputs={'label': 5})

    def test_work_calc_finish(self):
        p = test_utils.DummyProcess()
        self.assertFalse(p.calc.is_finished_ok)
        work.launch.run(p)
        self.assertTrue(p.calc.is_finished_ok)

    def test_calculation_input(self):
        @work.workfunction
        def simple_wf():
            return {'a': Int(6), 'b': Int(7)}

        outputs, pid = work.launch.run_get_pid(simple_wf)
        calc = load_node(pid)

        dp = test_utils.DummyProcess(inputs={'calc': calc})
        work.launch.run(dp)

        input_calc = dp.calc.get_inputs_dict()['calc']
        self.assertTrue(isinstance(input_calc, FrozenDict))
        self.assertEqual(input_calc['a'], outputs['a'])

    def test_save_instance_state(self):
        proc = test_utils.DummyProcess()
        # Save the instance state
        bundle = work.Bundle(proc)
        proc2 = bundle.unbundle()

    def test_process_type_with_entry_point(self):
        """
        For a process with a registered entry point, the process_type will be its formatted entry point string
        """
        from aiida.orm import CalculationFactory, Code

        code = Code()
        code.set_remote_computer_exec((self.computer, '/bin/true'))
        code.store()

        parameters = ParameterData(dict={})
        template = ParameterData(dict={})
        options = {
            'resources': {
                'num_machines': 1,
                'tot_num_mpiprocs': 1
            },
            'max_wallclock_seconds': 1,
        }

        inputs = {
            'code': code,
            'parameters': parameters,
            'template': template,
            'options': options,
        }

        entry_point = 'simpleplugins.templatereplacer'
        calculation = CalculationFactory(entry_point)
        job_process = calculation.process()
        process = job_process(inputs=inputs)

        expected_process_type = 'aiida.calculations:{}'.format(entry_point)
        self.assertEqual(process.calc.process_type, expected_process_type)

        # Verify that load_process_class on the calculation node returns the original entry point class
        recovered_process = process.calc.load_process_class()
        self.assertEqual(recovered_process, calculation)

    def test_process_type_without_entry_point(self):
        """
        For a process without a registered entry point, the process_type will fall back on the fully
        qualified class name
        """
        process = test_utils.DummyProcess()
        expected_process_type = '{}.{}'.format(process.__class__.__module__, process.__class__.__name__)
        self.assertEqual(process.calc.process_type, expected_process_type)

        # Verify that load_process_class on the calculation node returns the original entry point class
        recovered_process = process.calc.load_process_class()
        self.assertEqual(recovered_process, process.__class__)


class TestFunctionProcess(AiidaTestCase):

    def test_fixed_inputs(self):
        def wf(a, b, c):
            return {'a': a, 'b': b, 'c': c}

        inputs = {'a': Int(4), 'b': Int(5), 'c': Int(6)}
        function_process_class = work.FunctionProcess.build(wf)
        self.assertEqual(work.run(function_process_class, **inputs), inputs)

    def test_kwargs(self):
        def wf_with_kwargs(**kwargs):
            return kwargs

        def wf_without_kwargs():
            return Int(4)

        def wf_fixed_args(a):
            return {'a': a}

        a = Int(4)
        inputs = {'a': a}

        function_process_class = work.FunctionProcess.build(wf_with_kwargs)
        outs = work.launch.run(function_process_class, **inputs)
        self.assertEqual(outs, inputs)

        function_process_class = work.FunctionProcess.build(wf_without_kwargs)
        with self.assertRaises(ValueError):
            work.launch.run(function_process_class, **inputs)

        function_process_class = work.FunctionProcess.build(wf_fixed_args)
        outs = work.launch.run(function_process_class, **inputs)
        self.assertEqual(outs, inputs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Testing Session possible problems.
"""

import os
from sqlalchemy.orm import sessionmaker

import aiida.backends
from aiida.backends.testbase import AiidaTestCase
from aiida.common.utils import get_configured_user_email


class TestSessionSqla(AiidaTestCase):
    """
    The following tests check that the session works as expected in some
    problematic examples. When a session is initialized with
    expire_on_commit=False allows more permissive behaviour since committed
    objects that remain in the session do not need refresh. The opposite
    happens when expire_on_commit=True.

    Moreover, 2 ways of storing objects are tested, i.e. adding the objects
    manually to the session and committing it & by using the build-in store
    method of the ORM objects.
    """

    def set_connection(self, expire_on_commit=True):
        # Creating a sessionmaker with the desired parameters
        ## Note: to check if this is still correct with the new
        ## way of managing connections and sessions in SQLA...
        ## For instance, we should use probably a scopedsession wrapper
        Session = sessionmaker(expire_on_commit=expire_on_commit)
        aiida.backends.sqlalchemy.sessionfactory = Session(
            bind=self._AiidaTestCase__backend_instance.connection)

        # Cleaning the database
        self.clean_db()
        aiida.backends.sqlalchemy.get_scoped_session().expunge_all()

    def drop_connection(self):
        session = aiida.backends.sqlalchemy.get_scoped_session()
        session.expunge_all()
        session.close()
        aiida.backends.sqlalchemy.sessionfactory = None

    def test_session_update_and_expiration_1(self):
        """
        expire_on_commit=True & adding manually and committing
        computer and code objects.
        """
        from aiida.orm.implementation.sqlalchemy.computer import Computer
        from aiida.orm.implementation.sqlalchemy.code import Code

        self.set_connection(expire_on_commit=True)

        session = aiida.backends.sqlalchemy.get_scoped_session()

        user = self.backend.users.create(email=get_configured_user_email())
        session.add(user.dbuser)
        session.commit()

        defaults = dict(name='localhost',
                        hostname='localhost',
                        transport_type='local',
                        scheduler_type='pbspro',
                        workdir='/tmp/aiida')
        computer = Computer(**defaults)
        session.add(computer._dbcomputer)
        session.commit()

        code = Code()
        code.set_remote_computer_exec((computer, '/x.x'))
        session.add(code._dbnode)
        session.commit()

        self.drop_connection()

    def test_session_update_and_expiration_2(self):
        """
        expire_on_commit=True & committing computer and code objects with
        their built-in store function.
        """
        from aiida.backends.sqlalchemy.models.user import DbUser
        from aiida.orm.implementation.sqlalchemy.computer import Computer
        from aiida.orm.implementation.sqlalchemy.code import Code

        session = aiida.backends.sqlalchemy.get_scoped_session()

        self.set_connection(expire_on_commit=True)

        user = self.backend.users.create(email=get_configured_user_email())
        session.add(user.dbuser)
        session.commit()

        defaults = dict(name='localhost',
                        hostname='localhost',
                        transport_type='local',
                        scheduler_type='pbspro',
                        workdir='/tmp/aiida')
        computer = Computer(**defaults)
        computer.store()

        code = Code()
        code.set_remote_computer_exec((computer, '/x.x'))
        code.store()

        self.drop_connection()

    def test_session_update_and_expiration_3(self):
        """
        expire_on_commit=False & adding manually and committing
        computer and code objects.
        """
        from aiida.orm.implementation.sqlalchemy.computer import Computer
        from aiida.orm.implementation.sqlalchemy.code import Code

        self.set_connection(expire_on_commit=False)

        session = aiida.backends.sqlalchemy.get_scoped_session()

        user = self.backend.users.create(email=get_configured_user_email())
        session.add(user.dbuser)
        session.commit()

        defaults = dict(name='localhost',
                        hostname='localhost',
                        transport_type='local',
                        scheduler_type='pbspro',
                        workdir='/tmp/aiida')
        computer = Computer(**defaults)
        session.add(computer._dbcomputer)
        session.commit()

        code = Code()
        code.set_remote_computer_exec((computer, '/x.x'))
        session.add(code._dbnode)
        session.commit()

        self.drop_connection()

    def test_session_update_and_expiration_4(self):
        """
        expire_on_commit=False & committing computer and code objects with
        their built-in store function.
        """
        self.set_connection(expire_on_commit=False)

        from aiida.orm.implementation.sqlalchemy.computer import Computer
        from aiida.orm.implementation.sqlalchemy.code import Code

        session = aiida.backends.sqlalchemy.get_scoped_session()

        user = self.backend.users.create(email=get_configured_user_email())
        session.add(user.dbuser)
        session.commit()

        defaults = dict(name='localhost',
                        hostname='localhost',
                        transport_type='local',
                        scheduler_type='pbspro',
                        workdir='/tmp/aiida')
        computer = Computer(**defaults)
        computer.store()

        code = Code()
        code.set_remote_computer_exec((computer, '/x.x'))
        code.store()

        self.drop_connection()

    def test_session_wfdata(self):
        """
        This test checks that the
        aiida.backends.sqlalchemy.models.workflow.DbWorkflowData#set_value
        method works as expected. There were problems with the DbNode object
        that was added as a value to a DbWorkflowData object. If there was
        an older version of the dbnode in the session than the one given to
        to the DbWorkflowData#set_value then there was a collision in the
        session and SQLA identity map.
        """
        from aiida.orm.node import Node
        from aiida.workflows.test import WFTestSimpleWithSubWF
        from aiida.backends.sqlalchemy import get_scoped_session
        from aiida.orm.utils import load_node

        # Create a node and store it
        n = Node()
        n.store()

        # Keep some useful information
        n_id = n.id
        old_dbnode = n._dbnode

        # Get the session
        sess = get_scoped_session()
        # Remove everything from the session
        sess.expunge_all()

        # Create a workflow and store it
        wf = WFTestSimpleWithSubWF()
        wf.store()

        # Load a new version of the node
        n_reloaded = load_node(n_id)

        # Remove everything from the session
        sess.expunge_all()

        # Add the dbnode that was originally added to the session
        sess.add(old_dbnode)

        # Add as attribute the node that was added after the first cleanup
        # of the session
        # At this point the following command should not fail
        wf.add_attribute('a', n_reloaded)

    def test_node_access_with_sessions(self):
        """
        This checks that changes to a node from a different session (e.g. different interpreter,
        or the daemon) are immediately reflected on the AiiDA node when read directly e.g. a change
        to node.description will immediately be seen.

        Tests for bug #1372
        """
        from aiida.utils import timezone
        from aiida.orm.implementation.sqlalchemy.node import Node
        from aiida.orm.implementation.sqlalchemy.node import DbNode
        import aiida.backends.sqlalchemy as sa
        from sqlalchemy.orm import sessionmaker

        Session = sessionmaker(bind=sa.engine)
        custom_session = Session()

        node = Node().store()
        master_session = node._dbnode.session
        self.assertIsNot(master_session, custom_session)

        # Manually load the DbNode in a different session
        dbnode_reloaded = custom_session.query(DbNode).get(node.id)

        # Now, go through one by one changing the possible attributes (of the model)
        # and check that they're updated when the user reads them from the aiida node

        def check_attrs_match(name):
            node_attr = getattr(node, name)
            dbnode_attr = getattr(dbnode_reloaded, name)
            self.assertEqual(
                node_attr, dbnode_attr,
                "Values of '{}' don't match ({} != {})".format(name, node_attr, dbnode_attr))

        def do_value_checks(attr_name, original, changed):
            try:
                setattr(node, attr_name, original)
            except AttributeError:
                # This may mean that it is immutable, but we should still be able to
                # change it below directly through the dbnode
                pass
            # Refresh the custom session and make sure they match
            custom_session.refresh(dbnode_reloaded, attribute_names=[str_attr])
            check_attrs_match(attr_name)

            # Change the value in the custom session via the DbNode
            setattr(dbnode_reloaded, attr_name, changed)
            custom_session.commit()

            # Check that the Node 'sees' the change
            check_attrs_match(str_attr)

        for str_attr in ['label', 'description']:
            do_value_checks(str_attr, 'original', 'changed')

        do_value_checks('nodeversion', 1, 2)
        do_value_checks('public', True, False)

        for str_attr in ['ctime', 'mtime']:
            do_value_checks(str_attr, timezone.now(), timezone.now())

        # Attributes
        self.assertDictEqual(node._attributes(), dbnode_reloaded.attributes)
        dbnode_reloaded.attributes['test_attrs'] = 'Boo!'
        custom_session.commit()
        self.assertDictEqual(node._attributes(), dbnode_reloaded.attributes)

        # Extras
        self.assertDictEqual(node.get_extras(), dbnode_reloaded.extras)
        dbnode_reloaded.extras['test_extras'] = 'Boo!'
        custom_session.commit()
        self.assertDictEqual(node._attributes(), dbnode_reloaded.attributes)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=global-statement
"""Runners that can run and submit processes."""
from collections import namedtuple
from contextlib import contextmanager
import logging
import tornado.ioloop

import plumpy

from aiida.orm import load_node, load_workflow
from . import futures
from . import persistence
from . import rmq
from . import transports
from . import utils

__all__ = ['Runner', 'DaemonRunner', 'new_runner', 'set_runner', 'get_runner']

ResultAndNode = namedtuple('ResultAndNode', ['result', 'node'])
ResultAndPid = namedtuple('ResultAndPid', ['result', 'pid'])

RUNNER = None


def new_runner(**kwargs):
    """
    Create a default runner optionally passing keyword arguments

    :param kwargs: arguments to be passed to Runner constructor
    :return: a new runner instance
    """
    if 'rmq_config' not in kwargs:
        kwargs['rmq_config'] = rmq.get_rmq_config()

    return Runner(**kwargs)


def get_runner():
    """
    Get the global runner instance

    :returns: the global runner
    """
    global RUNNER
    if RUNNER is None or RUNNER.is_closed():
        RUNNER = new_runner()
    return RUNNER


def set_runner(runner):
    """
    Set the global runner instance

    :param runner: the runner instance to set as the global runner
    """
    global RUNNER
    RUNNER = runner


class Runner(object):
    """Class that can launch processes by running in the current interpreter or by submitting them to the daemon."""

    _persister = None
    _rmq_connector = None
    _communicator = None
    _closed = False

    # pylint: disable=too-many-arguments
    def __init__(self,
                 rmq_config=None,
                 poll_interval=0.,
                 loop=None,
                 rmq_submit=False,
                 enable_persistence=True,
                 persister=None):
        self._loop = loop if loop is not None else tornado.ioloop.IOLoop()
        self._poll_interval = poll_interval
        self._rmq_submit = rmq_submit
        self._transport = transports.TransportQueue(self._loop)

        if enable_persistence:
            self._persister = persister if persister is not None else persistence.AiiDAPersister()

        if rmq_config is not None:
            self._setup_rmq(**rmq_config)
        elif self._rmq_submit:
            logger = logging.getLogger(__name__)
            logger.warning('Disabling rmq submission, no RMQ config provided')
            self._rmq_submit = False

        # Save kwargs for creating child runners
        self._kwargs = {
            'rmq_config': rmq_config,
            'poll_interval': poll_interval,
            'rmq_submit': rmq_submit,
            'enable_persistence': enable_persistence
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def loop(self):
        return self._loop

    @property
    def rmq(self):
        return self._rmq

    @property
    def transport(self):
        return self._transport

    @property
    def persister(self):
        return self._persister

    @property
    def communicator(self):
        return self._communicator

    def is_closed(self):
        return self._closed

    def start(self):
        """Start the internal event loop."""
        self._loop.start()

    def stop(self):
        """Stop the internal event loop."""
        self._loop.stop()

    def run_until_complete(self, future):
        """Run the loop until the future has finished and return the result."""
        with utils.loop_scope(self._loop):
            return self._loop.run_sync(lambda: future)

    def close(self):
        """Close the runner by stopping the loop and disconnecting the RmqConnector if it has one."""
        assert not self._closed

        self.stop()

        if self._rmq_connector is not None:
            self._rmq_connector.disconnect()

        self._closed = True

    def instantiate_process(self, process, *args, **inputs):
        """
        Return an instance of the process with the given runner and inputs. The function can deal with various types
        of the `process`:

            * Process instance: will simply return the instance
            * JobCalculation class: will construct the JobProcess and instantiate it
            * ProcessBuilder instance: will instantiate the Process from the class and inputs defined within it
            * Process class: will instantiate with the specified inputs

        If anything else is passed, a ValueError will be raised

        :param self: instance of a Runner
        :param process: Process instance or class, JobCalculation class or ProcessBuilder instance
        :param inputs: the inputs for the process to be instantiated with
        """
        from aiida.orm.calculation.job import JobCalculation
        from aiida.work.process_builder import ProcessBuilder
        from aiida.work.processes import Process

        if isinstance(process, Process):
            assert not args
            assert not inputs
            assert self is process.runner
            return process

        if isinstance(process, ProcessBuilder):
            builder = process
            process_class = builder.process_class
            inputs.update(**builder)
        elif issubclass(process, JobCalculation):
            process_class = process.process()
        elif issubclass(process, Process):
            process_class = process
        else:
            raise ValueError('invalid process {}, needs to be Process, JobCalculation or ProcessBuilder'.format(
                type(process)))

        process = process_class(runner=self, inputs=inputs)

        return process

    def submit(self, process, *args, **inputs):
        """
        Submit the process with the supplied inputs to this runner immediately returning control to
        the interpreter. The return value will be the calculation node of the submitted process

        :param process: the process class to submit
        :param inputs: the inputs to be passed to the process
        :return: the calculation node of the process
        """
        assert not utils.is_workfunction(process), 'Cannot submit a workfunction'
        assert not self._closed

        process = self.instantiate_process(process, *args, **inputs)
        if self._rmq_submit:
            self.persister.save_checkpoint(process)
            process.close()
            self.rmq.continue_process(process.pid)
        else:
            # Run in this runner
            self.loop.add_callback(process.step_until_terminated)

        return process.calc

    def _run(self, process, *args, **inputs):
        """
        Run the process with the supplied inputs in this runner that will block until the process is completed.
        The return value will be the results of the completed process

        :param process: the process class or workfunction to run
        :param inputs: the inputs to be passed to the process
        :return: tuple of the outputs of the process and the calculation node
        """
        assert not self._closed

        if utils.is_workfunction(process):
            result, node = process.run_get_node(*args, **inputs)
            return result, node

        with utils.loop_scope(self.loop):
            process = self.instantiate_process(process, *args, **inputs)
            process.execute()
            return process.outputs, process.calc

    def run(self, process, *args, **inputs):
        """
        Run the process with the supplied inputs in this runner that will block until the process is completed.
        The return value will be the results of the completed process

        :param process: the process class or workfunction to run
        :param inputs: the inputs to be passed to the process
        :return: the outputs of the process
        """
        result, _ = self._run(process, *args, **inputs)
        return result

    def run_get_node(self, process, *args, **inputs):
        """
        Run the process with the supplied inputs in this runner that will block until the process is completed.
        The return value will be the results of the completed process

        :param process: the process class or workfunction to run
        :param inputs: the inputs to be passed to the process
        :return: tuple of the outputs of the process and the calculation node
        """
        result, node = self._run(process, *args, **inputs)
        return ResultAndNode(result, node)

    def run_get_pid(self, process, *args, **inputs):
        """
        Run the process with the supplied inputs in this runner that will block until the process is completed.
        The return value will be the results of the completed process

        :param process: the process class or workfunction to run
        :param inputs: the inputs to be passed to the process
        :return: tuple of the outputs of the process and process pid
        """
        result, node = self._run(process, *args, **inputs)
        return ResultAndPid(result, node.pk)

    def call_on_legacy_workflow_finish(self, pk, callback):
        """
        Callback to be called when the workflow of the given pk is terminated

        :param pk: the pk of the workflow
        :param callback: the function to be called upon workflow termination
        """
        workflow = load_workflow(pk=pk)
        self._poll_legacy_wf(workflow, callback)

    def call_on_calculation_finish(self, pk, callback):
        """
        Callback to be called when the calculation of the given pk is terminated

        :param pk: the pk of the calculation
        :param callback: the function to be called upon calculation termination
        """
        calculation = load_node(pk=pk)
        self._poll_calculation(calculation, callback)

    def get_calculation_future(self, pk):
        """
        Get a future for an orm Calculation. The future will have the calculation node
        as the result when finished.

        :return: A future representing the completion of the calculation node
        """
        return futures.CalculationFuture(pk, self._loop, self._poll_interval, self._communicator)

    @contextmanager
    def child_runner(self):
        """
        Contextmanager that will yield a runner that is a child of this runner

        :returns: a Runner instance that inherits the attributes of this runner
        """
        runner = self._create_child_runner()
        try:
            yield runner
        finally:
            runner.close()

    def _setup_rmq(self, url, prefix=None, task_prefetch_count=None, testing_mode=False):
        """
        Setup the RabbitMQ connection by creating a connector, communicator and control panel

        :param url: the url to use for the connector
        :param prefix: the rmq prefix to use for the communicator
        :param task_prefetch_count: the maximum number of tasks the communicator may retrieve at a given time
        :param testing_mode: whether to create a communicator in testing mode
        """
        self._rmq_connector = plumpy.rmq.RmqConnector(amqp_url=url, loop=self._loop)

        self._rmq_communicator = plumpy.rmq.RmqCommunicator(
            self._rmq_connector,
            exchange_name=rmq.get_message_exchange_name(prefix),
            task_queue=rmq.get_launch_queue_name(prefix),
            testing_mode=testing_mode,
            task_prefetch_count=task_prefetch_count)

        self._rmq = rmq.ProcessControlPanel(prefix=prefix, rmq_connector=self._rmq_connector, testing_mode=testing_mode)
        self._communicator = self._rmq.communicator

        # Establish RMQ connection
        self._communicator.connect()

    def _create_child_runner(self):
        return Runner(**self._kwargs)

    def _poll_legacy_wf(self, workflow, callback):
        if workflow.has_finished_ok() or workflow.has_failed():
            self._loop.add_callback(callback, workflow.pk)
        else:
            self._loop.call_later(self._poll_interval, self._poll_legacy_wf, workflow, callback)

    def _poll_calculation(self, calc_node, callback):
        if calc_node.is_terminated:
            self._loop.add_callback(callback, calc_node.pk)
        else:
            self._loop.call_later(self._poll_interval, self._poll_calculation, calc_node, callback)


class DaemonRunner(Runner):
    """
    A sub class of Runner suited for a daemon runner
    """

    def __init__(self, *args, **kwargs):
        kwargs['rmq_submit'] = True
        super(DaemonRunner, self).__init__(*args, **kwargs)

    def _setup_rmq(self, url, prefix=None, task_prefetch_count=None, testing_mode=False):
        super(DaemonRunner, self)._setup_rmq(url, prefix, task_prefetch_count, testing_mode)

        # Create a context for loading new processes
        load_context = plumpy.LoadSaveContext(runner=self)

        # Listen for incoming launch requests
        task_receiver = rmq.ProcessLauncher(
            loop=self.loop, persister=self.persister, load_context=load_context, loader=persistence.get_object_loader())
        self.communicator.add_task_subscriber(task_receiver)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import abc
import collections
import enum
import inspect
import itertools
import plumpy
import uuid
import traceback
from pika.exceptions import ConnectionClosed

from plumpy import ProcessState

from aiida.common import exceptions
from aiida.common.lang import override, protected
from aiida.common.links import LinkType
from aiida.common.log import LOG_LEVEL_REPORT
from aiida.orm import load_node
from aiida.orm.node import Node
from aiida.orm.calculation import Calculation
from aiida.orm.calculation.function import FunctionCalculation
from aiida.orm.calculation.work import WorkCalculation
from aiida.orm.data import Data
from aiida.utils.serialize import serialize_data, deserialize_data
from aiida.work.ports import InputPort, PortNamespace
from aiida.work.process_spec import ProcessSpec, ExitCode
from aiida.work.process_builder import ProcessBuilder
from .runners import get_runner
from . import utils

__all__ = ['Process', 'ProcessState', 'FunctionProcess']


@plumpy.auto_persist('_parent_pid', '_enable_persistence')
class Process(plumpy.Process):
    """
    This class represents an AiiDA process which can be executed and will
    have full provenance saved in the database.
    """
    __metaclass__ = abc.ABCMeta

    _spec_type = ProcessSpec
    _calc_class = WorkCalculation

    SINGLE_RETURN_LINKNAME = 'result'

    class SaveKeys(enum.Enum):
        """
        Keys used to identify things in the saved instance state bundle.
        """
        CALC_ID = 'calc_id'

    @classmethod
    def define(cls, spec):
        super(Process, cls).define(spec)
        spec.input('store_provenance', valid_type=bool, default=True, non_db=True)
        spec.input('description', valid_type=basestring, required=False, non_db=True)
        spec.input('label', valid_type=basestring, required=False, non_db=True)
        spec.inputs.valid_type = (Data, Calculation)
        spec.outputs.valid_type = (Data,)

    @classmethod
    def get_builder(cls):
        return ProcessBuilder(cls)

    @classmethod
    def get_or_create_db_record(cls):
        """
        Create a database calculation node that represents what happened in
        this process.
        :return: A calculation
        """
        return cls._calc_class()

    def __init__(self, inputs=None, logger=None, runner=None, parent_pid=None, enable_persistence=True):
        self._runner = runner if runner is not None else get_runner()

        super(Process, self).__init__(
            inputs=self.spec().inputs.serialize(inputs),
            logger=logger,
            loop=self._runner.loop,
            communicator=self.runner.communicator)

        self._calc = None
        self._parent_pid = parent_pid
        self._enable_persistence = enable_persistence
        if self._enable_persistence and self.runner.persister is None:
            self.logger.warning('Disabling persistence, runner does not have a persister')
            self._enable_persistence = False

    def on_create(self):
        super(Process, self).on_create()
        # If parent PID hasn't been supplied try to get it from the stack
        if self._parent_pid is None and Process.current():
            current = Process.current()
            if isinstance(current, Process):
                self._parent_pid = current._pid
        self._pid = self._create_and_setup_db_record()

    def init(self):
        super(Process, self).init()
        if self._logger is None:
            self.set_logger(self._calc.logger)

    def has_finished(self):
        """
        Has the process finished i.e. completed running normally, without abort
        or an exception.

        :return: True if finished, False otherwise
        :rtype: bool
        """
        return self.state == ProcessState.FINISHED

    @property
    def calc(self):
        return self._calc

    @override
    def save_instance_state(self, out_state, save_context):
        super(Process, self).save_instance_state(out_state, save_context)

        if self.inputs.store_provenance:
            assert self.calc.is_stored

        out_state[self.SaveKeys.CALC_ID.value] = self.pid

    def get_provenance_inputs_iterator(self):
        return itertools.ifilter(lambda kv: not kv[0].startswith('_'), self.inputs.iteritems())

    @override
    def load_instance_state(self, saved_state, load_context):
        if 'runner' in load_context:
            self._runner = load_context.runner
        else:
            self._runner = get_runner()

        load_context = load_context.copyextend(loop=self._runner.loop, communicator=self._runner.communicator)
        super(Process, self).load_instance_state(saved_state, load_context)

        if self.SaveKeys.CALC_ID.value in saved_state:
            self._calc = load_node(saved_state[self.SaveKeys.CALC_ID.value])
            self._pid = self.calc.pk
        else:
            self._pid = self._create_and_setup_db_record()

        self.calc.logger.info('Loaded process<{}> from saved state'.format(self.calc.pk))

    def kill(self, msg=None):
        """
        Kill the process and all the children calculations it called
        """
        self._calc.logger.info('Kill Process<{}>'.format(self._calc.pk))
        result = super(Process, self).kill(msg)

        # Only kill children if we could be killed ourselves
        if result is not False:
            killing = []
            for child in self.calc.called:
                try:
                    result = self.runner.rmq.kill_process(child.pk, 'Killed by parent<{}>'.format(self.calc.pk))
                    if isinstance(result, plumpy.Future):
                        killing.append(result)
                except ConnectionClosed:
                    self.logger.info('no connection available to kill child<{}>'.format(child.pk))

            if isinstance(result, plumpy.Future):
                # We ourselves are waiting to be killed so add it to the list
                killing.append(result)

            if killing:
                # We are waiting for things to be killed, so return the 'gathered' future
                result = plumpy.gather(result)

        return result

    @override
    def out(self, output_port, value=None):
        if value is None:
            # In this case assume that output_port is the actual value and there
            # is just one return value
            value = output_port
            output_port = self.SINGLE_RETURN_LINKNAME

        if isinstance(value, Node) and not value.is_stored:
            value.store()

        return super(Process, self).out(output_port, value)

    def out_many(self, out_dict):
        """
        Add all values given in ``out_dict`` to the outputs. The keys of the dictionary will be used as output names.
        """
        for key, value in out_dict.items():
            self.out(key, value)

    # region Process messages
    @override
    def on_entering(self, state):
        super(Process, self).on_entering(state)
        # Update the node attributes every time we enter a new state

    def on_entered(self, from_state):
        super(Process, self).on_entered(from_state)
        self.update_node_state(self._state)
        if self._enable_persistence and not self._state.is_terminal():
            try:
                self.runner.persister.save_checkpoint(self)
            except plumpy.PersistenceError:
                self.logger.warning(
                    "Exception trying to save checkpoint, this means you will "
                    "not be able to restart in case of a crash until the next successful checkpoint.")
                self.logger.debug(
                    "Exception trying to save checkpoint:\n{}".format(traceback.format_exc()))

        # Update the latest process state change timestamp
        utils.set_process_state_change_timestamp(self)

    @override
    def on_terminated(self):
        """
        Called when a Process enters a terminal state.
        """
        super(Process, self).on_terminated()
        if self._enable_persistence:
            try:
                self.runner.persister.delete_checkpoint(self.pid)
            except BaseException as exception:
                self.logger.warning('Failed to delete checkpoint: {}\n{}'.format(exception, traceback.format_exc()))

        try:
            self.calc.seal()
        except exceptions.ModificationNotAllowed:
            pass

    @override
    def on_except(self, exc_info):
        """
        Log the exception by calling the report method with formatted stack trace from exception info object
        and store the exception string as a node attribute

        :param exc_info: the sys.exc_info() object
        """
        super(Process, self).on_except(exc_info)
        self.calc._set_exception(''.join(traceback.format_exception(exc_info[0], exc_info[1], None)))
        self.report(''.join(traceback.format_exception(*exc_info)))

    @override
    def on_finish(self, result, successful):
        """
        Set the finish status on the Calculation node
        """
        super(Process, self).on_finish(result, successful)

        if result is None or isinstance(result, int):
            self.calc._set_exit_status(result)
        elif isinstance(result, ExitCode):
            self.calc._set_exit_status(result.status)
            self.calc._set_exit_message(result.message)
        else:
            raise ValueError('the result should be an integer, ExitCode or None, got {} {} {}'.format(type(result), result, self.pid))

    @override
    def on_output_emitting(self, output_port, value):
        """
        The process has emitted a value on the given output port.

        :param output_port: The output port name the value was emitted on
        :param value: The value emitted
        """
        super(Process, self).on_output_emitting(output_port, value)

        if not isinstance(value, Data):
            raise TypeError(
                'Values output from process must be instances of AiiDA Data types, got {}'.format(value.__class__)
            )

    # end region

    def run_process(self, process, *args, **inputs):
        with self.runner.child_runner() as runner:
            return runner.run(process, *args, **inputs)

    def submit(self, process, *args, **kwargs):
        return self.runner.submit(process, *args, **kwargs)

    @property
    def runner(self):
        return self._runner

    @protected
    def get_parent_calc(self):
        # Can't get it if we don't know our parent
        if self._parent_pid is None:
            return None

        return load_node(pk=self._parent_pid)

    @protected
    def report(self, msg, *args, **kwargs):
        """
        Log a message to the logger, which should get saved to the
        database through the attached DbLogHandler. The class name and function
        name of the caller are prepended to the given message
        """
        message = '[{}|{}|{}]: {}'.format(self.calc.pk, self.__class__.__name__, inspect.stack()[1][3], msg)
        self.logger.log(LOG_LEVEL_REPORT, message, *args, **kwargs)

    def _create_and_setup_db_record(self):
        self._calc = self.get_or_create_db_record()
        self._setup_db_record()
        if self.inputs.store_provenance:
            try:
                self.calc.store_all()
                if self.calc.is_finished_ok:
                    self._state = ProcessState.FINISHED
                    for name, value in self.calc.get_outputs_dict(link_type=LinkType.RETURN).items():
                        if name.endswith('_{pk}'.format(pk=value.pk)):
                            continue
                        self.out(name, value)
                    # This is needed for JobProcess. In that case, the outputs are
                    # returned regardless of whether they end in '_pk'
                    for name, value in self.calc.get_outputs_dict(link_type=LinkType.CREATE).items():
                        self.out(name, value)
            except exceptions.ModificationNotAllowed:
                # The calculation was already stored
                pass

        if self.calc.pk is not None:
            return self.calc.pk
        else:
            return uuid.UUID(self.calc.uuid)

    @override
    def encode_input_args(self, inputs):
        """
        Encode input arguments such that they may be saved in a Bundle

        :param inputs: A mapping of the inputs as passed to the process
        :return: The encoded inputs
        """
        return serialize_data(inputs)

    @override
    def decode_input_args(self, encoded):
        """
        Decode saved input arguments as they came from the saved instance state Bundle

        :param encoded:
        :return: The decoded input args
        """
        return deserialize_data(encoded)

    def update_node_state(self, state):
        self.update_outputs()
        self.calc._set_process_state(state.LABEL)

    def update_outputs(self):
        # Link up any new outputs
        new_outputs = set(self.outputs.keys()) - set(self.calc.get_outputs_dict(link_type=LinkType.RETURN).keys())
        for label in new_outputs:
            value = self.outputs[label]
            # Try making us the creator
            try:
                value.add_link_from(self.calc, label, LinkType.CREATE)
            except ValueError:
                # Must have already been created...nae dramas
                pass

            value.store()

            if utils.is_work_calc_type(self.calc):
                value.add_link_from(self.calc, label, LinkType.RETURN)

    def _setup_db_record(self):
        assert self.inputs is not None
        assert not self.calc.is_sealed, \
            "Calculation cannot be sealed when setting up the database record"

        # Save the name of this process
        self.calc._set_process_state(None)
        self.calc._set_process_label(self.__class__.__name__)
        self.calc._set_process_type(self.__class__)

        parent_calc = self.get_parent_calc()

        for name, input_value in self._flat_inputs().iteritems():

            if isinstance(input_value, Calculation):
                input_value = utils.get_or_create_output_group(input_value)

            if not input_value.is_stored:
                # If the input isn't stored then assume our parent created it
                if parent_calc:
                    input_value.add_link_from(parent_calc, "CREATE", link_type=LinkType.CREATE)
                if self.inputs.store_provenance:
                    input_value.store()

            self.calc.add_link_from(input_value, name)

        if parent_calc:
            self.calc.add_link_from(parent_calc, "CALL", link_type=LinkType.CALL)

        self._add_description_and_label()

    def _add_description_and_label(self):
        if self.inputs:
            description = self.inputs.get('description', None)
            if description is not None:
                self._calc.description = description
            label = self.inputs.get('label', None)
            if label is not None:
                self._calc.label = label

    def _flat_inputs(self):
        """
        Return a flattened version of the parsed inputs dictionary. The eventual
        keys will be a concatenation of the nested keys

        :return: flat dictionary of parsed inputs
        """
        return dict(self._flatten_inputs(self.spec().inputs, self.inputs))

    def _flatten_inputs(self, port, port_value, parent_name='', separator='_'):
        """
        Function that will recursively flatten the inputs dictionary, omitting inputs for ports that
        are marked as being non database storable

        :param port: port against which to map the port value, can be InputPort or PortNamespace
        :param port_value: value for the current port, can be a Mapping
        :param parent_name: the parent key with which to prefix the keys
        :param separator: character to use for the concatenation of keys
        """
        if (
                (port is None and isinstance(port_value, Node)) or
                (isinstance(port, InputPort) and not getattr(port, 'non_db', False))
        ):
            return [(parent_name, port_value)]
        elif (
                (port is None and isinstance(port_value, collections.Mapping)) or
                isinstance(port, PortNamespace)
        ):
            items = []
            for name, value in port_value.items():
                prefixed_key = parent_name + separator + name if parent_name else name

                try:
                    nested_port = port[name]
                except (KeyError, TypeError):
                    nested_port = None

                sub_items = self._flatten_inputs(
                    port=nested_port,
                    port_value=value,
                    parent_name=prefixed_key,
                    separator=separator
                )
                items.extend(sub_items)
            return items

        else:
            assert (port is None) or (isinstance(port, InputPort) and port.non_db)
            return []

    def exposed_inputs(self, process_class, namespace=None, agglomerate=True):
        """
        Gather a dictionary of the inputs that were exposed for a given Process class under an optional namespace.

        :param process_class: Process class whose inputs to try and retrieve

        :param namespace: PortNamespace in which to look for the inputs
        :type namespace: str

        :param agglomerate: If set to true, all parent namespaces of the given ``namespace`` will also be
            searched for inputs. Inputs in lower-lying namespaces take precedence.
        :type agglomerate: bool
        """
        exposed_inputs = {}

        namespace_list = self._get_namespace_list(namespace=namespace, agglomerate=agglomerate)
        for namespace in namespace_list:

            # The namespace None indicates the base level namespace
            if namespace is None:
                inputs = self.inputs
                port_namespace = self.spec().inputs
            else:
                inputs = self.inputs
                for ns in namespace.split('.'):
                    inputs = inputs[ns]
                try:
                    port_namespace = self.spec().inputs.get_port(namespace)
                except KeyError:
                    raise ValueError('this process does not contain the "{}" input namespace'.format(namespace))

            # Get the list of ports that were exposed for the given Process class in the current namespace
            exposed_inputs_list = self.spec()._exposed_inputs[namespace][process_class]

            for name, port in port_namespace.ports.iteritems():
                if name in inputs and name in exposed_inputs_list:
                    exposed_inputs[name] = inputs[name]

        return exposed_inputs

    def exposed_outputs(self, process_instance, process_class, namespace=None, agglomerate=True):
        """
        Gather the outputs which were exposed from the ``process_class`` and emitted by the specific
        ``process_instance`` in a dictionary.

        :param namespace: Namespace in which to search for exposed outputs.
        :type namespace: str

        :param agglomerate: If set to true, all parent namespaces of the given ``namespace`` will also
            be searched for outputs. Outputs in lower-lying namespaces take precedence.
        :type agglomerate: bool
        """
        namespace_separator = self.spec().namespace_separator

        output_key_map = {}
        # maps the exposed name to all outputs that belong to it
        top_namespace_map = collections.defaultdict(list)
        process_outputs_dict = {
            k: v for k, v in process_instance.get_outputs(also_labels=True, link_type=LinkType.RETURN)
        }

        for port_name in process_outputs_dict:
            top_namespace = port_name.split(namespace_separator)[0]
            top_namespace_map[top_namespace].append(port_name)

        for ns in self._get_namespace_list(namespace=namespace, agglomerate=agglomerate):
            # only the top-level key is stored in _exposed_outputs
            for top_name in top_namespace_map:
                if top_name in self.spec()._exposed_outputs[ns][process_class]:
                    output_key_map[top_name] = ns

        result = {}

        for top_name, ns in output_key_map.items():
            # collect all outputs belonging to the given top_name
            for port_name in top_namespace_map[top_name]:
                if ns is None:
                    result[port_name] = process_outputs_dict[port_name]
                else:
                    result[ns + namespace_separator + port_name] = process_outputs_dict[port_name]
        return result

    @staticmethod
    def _get_namespace_list(namespace=None, agglomerate=True):
        if not agglomerate:
            return [namespace]
        else:
            namespace_list = [None]
            if namespace is not None:
                split_ns = namespace.split('.')
                namespace_list.extend([
                    '.'.join(split_ns[:i])
                    for i in range(1, len(split_ns) + 1)
                ])
            return namespace_list


class FunctionProcess(Process):
    _func_args = None
    _calc_node_class = FunctionCalculation

    @staticmethod
    def _func(*args, **kwargs):
        """
        This is used internally to store the actual function that is being
        wrapped and will be replaced by the build method.
        """
        return {}

    @staticmethod
    def build(func, calc_node_class=None):
        """
        Build a Process from the given function.  All function arguments will
        be assigned as process inputs. If keyword arguments are specified then
        these will also become inputs.

        :param func: The function to build a process from
        :param calc_node_class: Provide a custom calculation class to be used,
            has to be constructable with no arguments
        :type calc_node_class: :class:`aiida.orm.calculation.Calculation`
        :return: A Process class that represents the function
        :rtype: :class:`FunctionProcess`
        """
        args, varargs, keywords, defaults = inspect.getargspec(func)
        nargs = len(args)
        ndefaults = len(defaults) if defaults else 0
        first_default_pos = nargs - ndefaults

        if calc_node_class is None:
            calc_node_class = FunctionCalculation

        if varargs is not None:
            raise ValueError('variadic arguments are not supported')

        def _define(cls, spec):
            super(FunctionProcess, cls).define(spec)

            for i in range(len(args)):
                default = ()
                if i >= first_default_pos:
                    default = defaults[i - first_default_pos]

                if spec.has_input(args[i]):
                    spec.inputs[args[i]].default = default
                else:
                    spec.input(args[i], valid_type=Data, default=default)

            # If the function support kwargs then allow dynamic inputs, otherwise disallow
            if keywords is not None:
                spec.inputs.dynamic = True
            else:
                spec.inputs.dynamic = False

            # Workfunctions return data types
            spec.outputs.valid_type = Data

        return type(
            func.__name__, (FunctionProcess,),
            {
                '_func': staticmethod(func),
                Process.define.__name__: classmethod(_define),
                '_func_args': args,
                '_calc_node_class': calc_node_class
            }
        )

    @classmethod
    def create_inputs(cls, *args, **kwargs):
        ins = {}
        if kwargs:
            ins.update(kwargs)
        if args:
            ins.update(cls.args_to_dict(*args))
        return ins

    @classmethod
    def args_to_dict(cls, *args):
        """
        Create an input dictionary (i.e. label: value) from supplied args.

        :param args: The values to use
        :return: A label: value dictionary
        """
        return dict(zip(cls._func_args, args))

    @classmethod
    def get_or_create_db_record(cls):
        return cls._calc_node_class()

    def __init__(self, *args, **kwargs):
        if kwargs.get('enable_persistence', False):
            raise RuntimeError('Cannot persist a workfunction')
        super(FunctionProcess, self).__init__(enable_persistence=False, *args, **kwargs)

    def execute(self):
        result = super(FunctionProcess, self).execute()
        # Create a special case for Process functions: They can return
        # a single value in which case you get this a not a dict
        if len(result) == 1 and self.SINGLE_RETURN_LINKNAME in result:
            return result[self.SINGLE_RETURN_LINKNAME]
        else:
            return result

    @override
    def _setup_db_record(self):
        super(FunctionProcess, self)._setup_db_record()
        self.calc.store_source_info(self._func)
        self.calc._set_process_label(self._func.__name__)

    @override
    def _run(self):
        args = []

        # Split the inputs into positional and keyword arguments
        args = [None] * len(self._func_args)
        kwargs = {}
        for name, value in self.inputs.items():
            try:
                if self.spec().inputs[name].non_db:
                    # Don't consider non-database inputs
                    continue
            except KeyError:
                pass  # No port found

            # Check if it is a positional arg, if not then keyword
            try:
                args[self._func_args.index(name)] = value
            except ValueError:
                kwargs[name] = value

        result = self._func(*args, **kwargs)

        if result is None or isinstance(result, ExitCode):
            return result

        if isinstance(result, Data):
            self.out(self.SINGLE_RETURN_LINKNAME, result)
        elif isinstance(result, collections.Mapping):
            for name, value in result.iteritems():
                self.out(name, value)
        else:
            raise TypeError(
                "Workfunction returned unsupported type '{}'\n"
                "Must be a Data type or a Mapping of {{string: Data}}".format(result.__class__))

        return ExitCode()
